@@ -13,7 +13,11 @@
 
 #include <ftw.h>
 #include <sys/stat.h>
+#ifdef __linux
 #include <sys/vfs.h>
+#elif defined IOS
+#import <Foundation/Foundation.h>
+#endif
 
 #include <chrono>
 #include <cstdio>
@@ -186,16 +190,17 @@ namespace
 
 namespace FileUtil
 {
+#ifndef MOBILEAPP
     void registerFileSystemForDiskSpaceChecks(const std::string& path)
     {
-        std::lock_guard<std::mutex> lock(fsmutex);
-
-        if (!path.empty())
+        const std::string::size_type lastSlash = path.rfind('/');
+        assert(path.empty() || lastSlash != std::string::npos);
+        if (lastSlash != std::string::npos)
         {
-            std::string dirPath = path;
-            std::string::size_type lastSlash = dirPath.rfind('/');
-            assert(lastSlash != std::string::npos);
-            dirPath = dirPath.substr(0, lastSlash + 1) + '.';
+            const std::string dirPath = path.substr(0, lastSlash + 1) + '.';
+            LOG_INF("Registering filesystem for space checks: [" << dirPath << "]");
+
+            std::lock_guard<std::mutex> lock(fsmutex);
 
             struct stat s;
             if (stat(dirPath.c_str(), &s) == 0)
@@ -207,19 +212,19 @@ namespace FileUtil
 
     std::string checkDiskSpaceOnRegisteredFileSystems(const bool cacheLastCheck)
     {
-        std::lock_guard<std::mutex> lock(fsmutex);
-
         static std::chrono::steady_clock::time_point lastCheck;
         std::chrono::steady_clock::time_point now(std::chrono::steady_clock::now());
 
-        // Don't check more often that once a minute
+        std::lock_guard<std::mutex> lock(fsmutex);
+
+        // Don't check more often than once a minute
         if (std::chrono::duration_cast<std::chrono::seconds>(now - lastCheck).count() < 60)
             return std::string();
 
         if (cacheLastCheck)
             lastCheck = now;
 
-        for (auto& i: filesystems)
+        for (const auto& i: filesystems)
         {
             if (!checkDiskSpace(i.path))
             {
@@ -229,27 +234,47 @@ namespace FileUtil
 
         return std::string();
     }
+#endif
 
     bool checkDiskSpace(const std::string& path)
     {
         assert(!path.empty());
 
+#ifndef MOBILEAPP
         bool hookResult;
         if (UnitBase::get().filterCheckDiskSpace(path, hookResult))
             return hookResult;
-
-        struct statfs sfs;
-        if (statfs(path.c_str(), &sfs) == -1)
-            return true;
+#endif
 
         // we should be able to run just OK with 5GB
         constexpr int64_t ENOUGH_SPACE = int64_t(5)*1024*1024*1024;
 
-        if (static_cast<int64_t>(sfs.f_bavail) * sfs.f_bsize > ENOUGH_SPACE)
+#ifdef __linux
+        struct statfs sfs;
+        if (statfs(path.c_str(), &sfs) == -1)
+            return true;
+
+        const int64_t freeBytes = static_cast<int64_t>(sfs.f_bavail) * sfs.f_bsize;
+
+        LOG_INF("Filesystem [" << path << "] has " << (freeBytes / 1024 / 1024) <<
+                " MB free (" << (sfs.f_bavail * 100. / sfs.f_blocks) << "%).");
+
+        if (freeBytes > ENOUGH_SPACE)
             return true;
 
         if (static_cast<double>(sfs.f_bavail) / sfs.f_blocks <= 0.05)
             return false;
+#elif defined IOS
+        NSDictionary *atDict = [[NSFileManager defaultManager] attributesOfFileSystemForPath:@"/" error:NULL];
+        long long freeSpace = [[atDict objectForKey:NSFileSystemFreeSize] longLongValue];
+        long long totalSpace = [[atDict objectForKey:NSFileSystemSize] longLongValue];
+
+        if (freeSpace > ENOUGH_SPACE)
+            return true;
+
+        if (static_cast<double>(freeSpace) / totalSpace <= 0.05)
+            return false;
+#endif
 
         return true;
     }
