@@ -7,7 +7,7 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  */
 
-#include "config.h"
+#include <config.h>
 
 #include "DummyLibreOfficeKit.hpp"
 
@@ -106,6 +106,7 @@ static void doc_setClientZoom(LibreOfficeKitDocument* pThis,
                                     int nTileTwipWidth,
                                     int nTileTwipHeight);
 static void doc_setClientVisibleArea(LibreOfficeKitDocument* pThis, int nX, int nY, int nWidth, int nHeight);
+static void doc_setOutlineState(LibreOfficeKitDocument* pThis, bool bColumn, int nLevel, int nIndex, bool bHidden);
 static int doc_createView(LibreOfficeKitDocument* pThis);
 static void doc_destroyView(LibreOfficeKitDocument* pThis, int nId);
 static void doc_setView(LibreOfficeKitDocument* pThis, int nId);
@@ -153,6 +154,7 @@ LibLODocument_Impl::LibLODocument_Impl()
         m_pDocumentClass->getCommandValues = doc_getCommandValues;
         m_pDocumentClass->setClientZoom = doc_setClientZoom;
         m_pDocumentClass->setClientVisibleArea = doc_setClientVisibleArea;
+        m_pDocumentClass->setOutlineState = doc_setOutlineState;
 
         m_pDocumentClass->createView = doc_createView;
         m_pDocumentClass->destroyView = doc_destroyView;
@@ -277,8 +279,7 @@ static char* doc_getPartName(LibreOfficeKitDocument* pThis, int nPart)
     (void) pThis;
     (void) nPart;
 
-    char* pMemory = static_cast<char*>(malloc(11));
-    strncpy(pMemory, "Dummy part", 10);
+    char* pMemory = strdup("Dummy part");
     return pMemory;
 
 }
@@ -400,13 +401,11 @@ static char* doc_getTextSelection(LibreOfficeKitDocument* pThis, const char* pMi
     (void) pMimeType;
     (void) pUsedMimeType;
 
-    char* pMemory = static_cast<char*>(malloc(11));
-    strncpy(pMemory, "Dummy text", 10);
+    char* pMemory = strdup("Dummy text");
 
     if (pUsedMimeType)
     {
-        *pUsedMimeType = static_cast<char*>(malloc(25));
-        strncpy(*pUsedMimeType, "text/plain;charset=utf-8", 24);
+        *pUsedMimeType = strdup("text/plain;charset=utf-8");
     }
 
     return pMemory;
@@ -440,8 +439,7 @@ static char* doc_getCommandValues(LibreOfficeKitDocument* pThis, const char* pCo
     (void) pThis;
     (void) pCommand;
 
-    char* pMemory = static_cast<char*>(malloc(1));
-    strncpy(pMemory, "", 0);
+    char* pMemory = strdup("");
     return pMemory;
 }
 
@@ -464,6 +462,15 @@ static void doc_setClientVisibleArea(LibreOfficeKitDocument* pThis, int nX, int
     (void) nHeight;
 }
 
+static void doc_setOutlineState(LibreOfficeKitDocument* pThis, bool bColumn, int nLevel, int nIndex, bool bHidden)
+{
+    (void) pThis;
+    (void) bColumn;
+    (void) nLevel;
+    (void) nIndex;
+    (void) bHidden;
+}
+
 static int doc_createView(LibreOfficeKitDocument* /*pThis*/)
 {
     return 1;
@@ -516,8 +523,7 @@ static char* lo_getError (LibreOfficeKit *pThis)
 {
     (void) pThis;
 
-    char* pMemory = static_cast<char*>(malloc(12));
-    strncpy(pMemory, "Dummy error", 11);
+    char* pMemory = strdup("Dummy error");
     return pMemory;
 }
 
@@ -558,8 +564,7 @@ static char* lo_getVersionInfo(LibreOfficeKit* /*pThis*/)
         "\"BuildId\": \"1\" "
         "}";
 
-    char* pVersion = static_cast<char*>(malloc(sizeof(version)));
-    strncpy(pVersion, version, sizeof(version) - 1);
+    char* pVersion = strdup(version);
     return pVersion;
 }
 
