@@ -1,3 +1,4 @@
+/* -*- js-indent-level: 8 -*- */
 /*
  * L.Map.Keyboard is handling keyboard interaction with the map, enabled by default.
  */
@@ -172,32 +173,25 @@ L.Map.Keyboard = L.Handler.extend({
 
 		this._map.on('mousedown', this._onMouseDown, this);
 		this._map.on('keydown keyup keypress', this._onKeyDown, this);
-		this._map.on('compositionstart compositionupdate compositionend textInput', this._onKeyDown, this);
+		this._map.on('compositionstart compositionupdate compositionend textInput', this._onIME, this);
 	},
 
 	removeHooks: function () {
 		this._map.off('mousedown', this._onMouseDown, this);
 		this._map.off('keydown keyup keypress', this._onKeyDown, this);
-		this._map.off('compositionstart compositionupdate compositionend textInput', this._onKeyDown, this);
+		this._map.off('compositionstart compositionupdate compositionend textInput', this._onIME, this);
 	},
 
-	/*
-	 * Returns true whenever the key event shall be ignored.
-	 * This means shift+insert and shift+delete (or "insert or delete when holding
-	 * shift down"). Those events are handled elsewhere to trigger "cut" and 
-	 * "paste" events, and need to be ignored in order to avoid double-handling them.
-	 */
 	_ignoreKeyEvent: function(e) {
-		var shift = e.originalEvent.shiftKey;
-		if ('key' in e.originalEvent) {
-			var key = e.originalEvent.key;
-			return (shift && (key === 'Delete' || key === 'Insert'));
-		} else {
-			// keyCode is not reliable in AZERTY/DVORAK keyboard layouts, is used
-			// only as a fallback for MSIE8.
-			var keyCode = e.originalEvent.keyCode;
-			return (shift && (keyCode === 45 || keyCode === 46));
+		var shift = e.originalEvent.shiftKey ? this.keyModifier.shift : 0;
+		if (shift && (e.originalEvent.keyCode === 45 || e.originalEvent.keyCode === 46)) {
+			// don't handle shift+insert, shift+delete
+			// These are converted to 'cut', 'paste' events which are
+			// automatically handled by us, so avoid double-handling
+			return true;
 		}
+
+		return false;
 	},
 
 	_setPanOffset: function (pan) {
@@ -258,7 +252,7 @@ L.Map.Keyboard = L.Handler.extend({
 			compEventFn = L.bind(docLayer._postCompositionEvent, docLayer, 0 /* winid */);
 		}
 		if (!inputEle) {
-			inputEle = this._map._clipboardContainer._textArea;
+			inputEle = this._map._clipboardContainer.activeElement();
 		}
 		this.modifier = 0;
 		var shift = e.originalEvent.shiftKey ? this.keyModifier.shift : 0;
@@ -302,31 +296,20 @@ L.Map.Keyboard = L.Handler.extend({
 		var charCode = e.originalEvent.charCode;
 		var keyCode = e.originalEvent.keyCode;
 
-		if (e.type === 'compositionstart' || e.type === 'compositionupdate') {
-			this._isComposing = true; // we are starting composing with IME
-			var txt = '';
-			for (var i = 0; i < e.originalEvent.data.length; i++) {
-				txt += e.originalEvent.data[i];
-			}
-			if (txt) {
-				compEventFn('input', txt);
-			}
-		}
-
-		if (e.type === 'compositionend') {
-			this._isComposing = false; // stop of composing with IME
-			// get the composited char codes
-			// clear the input now - best to do this ASAP so the input
-			// is clear for the next word
-			inputEle.value = '';
-		}
-
 		if (!this._isComposing && e.type === 'keyup') {
 			// not compositing and keyup, clear the input so it is ready
 			// for next word (or char only)
 			inputEle.value = '';
 		}
 
+		if ((this.modifier == this.keyModifier.alt || this.modifier == this.keyModifier.shift + this.keyModifier.alt) &&
+		    keyCode >= 48) {
+			// Presumably a Mac or iOS client accessing a "special character". Just ignore the alt modifier.
+			// But don't ignore it for Alt + non-printing keys.
+			this.modifier -= alt;
+			alt = 0;
+		}
+
 		var unoKeyCode = this._toUNOKeyCode(keyCode);
 
 		if (this.modifier) {
@@ -346,14 +329,12 @@ L.Map.Keyboard = L.Handler.extend({
 			}
 			else if (e.type === 'keydown') {
 				this._keyHandled = false;
-				this._bufferedTextInputEvent = null;
 
 				if (this.handleOnKeyDownKeys[keyCode] && charCode === 0) {
 					keyEventFn('input', charCode, unoKeyCode);
 				}
 			}
-			else if ((e.type === 'keypress' || e.type === 'compositionend') &&
-			         (!this.handleOnKeyDownKeys[keyCode] || charCode !== 0)) {
+			else if ((e.type === 'keypress') && (!this.handleOnKeyDownKeys[keyCode] || charCode !== 0)) {
 				if (charCode === keyCode && charCode !== 13) {
 					// Chrome sets keyCode = charCode for printable keys
 					// while LO requires it to be 0
@@ -364,47 +345,13 @@ L.Map.Keyboard = L.Handler.extend({
 					// key press times will be paired with the invalidation messages
 					docLayer._debugKeypressQueue.push(+new Date());
 				}
-				if (e.type === 'compositionend') {
-					// Set all keycodes to zero
-					compEventFn('end', '');
-				} else {
-					keyEventFn('input', charCode, unoKeyCode);
-				}
 
+				keyEventFn('input', charCode, unoKeyCode);
 				this._keyHandled = true;
 			}
-			else if (e.type === 'textInput') {
-				// Store the textInput event
-				this._bufferedTextInputEvent = e;
-			}
 			else if (e.type === 'keyup') {
-				// Hack for making space and spell-check text insert work
-				// in Chrome (on Andorid) or Chrome with IME.
-				//
-				// Chrome (Android) IME triggers keyup/keydown input with
-				// code 229 when hitting space (as with all composiiton events)
-				// with addition to 'textinput' event, in which we only see that
-				// space was entered. Similar situation is also when inserting
-				// a soft-keyboard spell-check item - it is visible only with
-				// 'textinput' event (no composition event is fired).
-				// To make this work we need to insert textinput.data here..
-				//
-				// TODO: Maybe make sure this is only triggered when keydown has
-				// 229 code. Also we need to detect that composition was overriden
-				// (part or whole word deleted) with the spell-checked word. (for
-				// example: enter 'tar' and with spell-check correct that to 'rat')
-
-				if (!this._keyHandled && this._bufferedTextInputEvent) {
-					var textInputData = this._bufferedTextInputEvent.originalEvent.data;
-					charCode = e.originalEvent.keyCode;
-					for (var idx = 0; idx < textInputData.length; idx++) {
-						keyEventFn('input', textInputData[idx].charCodeAt(), 0);
-					}
-				}
 				keyEventFn('up', charCode, unoKeyCode);
-
 				this._keyHandled = true;
-				this._bufferedTextInputEvent = null;
 			}
 			if (keyCode === 9) {
 				// tab would change focus to other DOM elements
@@ -437,6 +384,51 @@ L.Map.Keyboard = L.Handler.extend({
 		L.DomEvent.stopPropagation(e.originalEvent);
 	},
 
+	_onIME: function (e) {
+		if (e.type === 'compositionstart' || e.type === 'compositionupdate') {
+			this._isComposing = true; // we are starting composing with IME
+			var txt = '';
+			for (var i = 0; i < e.originalEvent.data.length; i++) {
+				txt += e.originalEvent.data[i];
+			}
+			if (txt) {
+				this._map._docLayer._postCompositionEvent(0, 'input', txt);
+			}
+		}
+
+		if (e.type === 'compositionend') {
+			this._isComposing = false; // stop of composing with IME
+			// get the composited char codes
+			// clear the input now - best to do this ASAP so the input
+			// is clear for the next word
+			this._map._clipboardContainer.setValue('');
+			// Set all keycodes to zero
+			this._map._docLayer._postCompositionEvent(0, 'end', '');
+		}
+
+		if (e.type === 'textInput' && !this._keyHandled) {
+			// Hack for making space and spell-check text insert work
+			// in Chrome (on Andorid) or Chrome with IME.
+			//
+			// Chrome (Android) IME triggers keyup/keydown input with
+			// code 229 when hitting space (as with all composiiton events)
+			// with addition to 'textinput' event, in which we only see that
+			// space was entered. Similar situation is also when inserting
+			// a soft-keyboard spell-check item - it is visible only with
+			// 'textinput' event (no composition event is fired).
+			// To make this work we need to insert textinput.data here..
+			//
+			// TODO: Maybe make sure this is only triggered when keydown has
+			// 229 code. Also we need to detect that composition was overriden
+			// (part or whole word deleted) with the spell-checked word. (for
+			// example: enter 'tar' and with spell-check correct that to 'rat')
+			var data = e.originalEvent.data;
+			for (var idx = 0; idx < data.length; idx++) {
+				this._map._docLayer._postKeyboardEvent('input', data[idx].charCodeAt(), 0);
+			}
+		}
+	},
+
 	_handleCtrlCommand: function (e) {
 		// Control
 		if (e.originalEvent.keyCode == 17)
