@@ -9,9 +9,13 @@
 
 #include <config.h>
 
-#include <stdio.h>
+#include "Socket.hpp"
+
+#include <cstring>
 #include <ctype.h>
 #include <iomanip>
+#include <stdio.h>
+#include <unistd.h>
 #include <zlib.h>
 
 #include <Poco/DateTime.h>
@@ -23,9 +27,10 @@
 #include <Poco/URI.h>
 
 #include <SigUtil.hpp>
-#include "Socket.hpp"
 #include "ServerSocket.hpp"
+#ifndef MOBILEAPP
 #include "SslSocket.hpp"
+#endif
 #include "WebSocketHandler.hpp"
 
 int SocketPoll::DefaultPollTimeoutMs = 5000;
@@ -34,8 +39,12 @@ std::atomic<bool> Socket::InhibitThreadChecks(false);
 
 int Socket::createSocket(Socket::Type type)
 {
+#ifndef MOBILEAPP
     int domain = type == Type::IPv4 ? AF_INET : AF_INET6;
     return socket(domain, SOCK_STREAM | SOCK_NONBLOCK, 0);
+#else
+    return fakeSocketSocket();
+#endif
 }
 
 // help with initialization order
@@ -57,10 +66,17 @@ SocketPoll::SocketPoll(const std::string& threadName)
       _stop(false),
       _threadStarted(false),
       _threadFinished(false),
+      _runOnClientThread(false),
       _owner(std::this_thread::get_id())
 {
     // Create the wakeup fd.
-    if (::pipe2(_wakeup, O_CLOEXEC | O_NONBLOCK) == -1)
+    if (
+#ifndef MOBILEAPP
+        ::pipe2(_wakeup, O_CLOEXEC | O_NONBLOCK) == -1
+#else
+        fakeSocketPipe2(_wakeup) == -1
+#endif
+        )
     {
         throw std::runtime_error("Failed to allocate pipe for SocketPoll [" + threadName + "] waking.");
     }
@@ -83,20 +99,30 @@ SocketPoll::~SocketPoll()
             getWakeupsArray().erase(it);
     }
 
+#ifndef MOBILEAPP
     ::close(_wakeup[0]);
     ::close(_wakeup[1]);
+#else
+    fakeSocketClose(_wakeup[0]);
+    fakeSocketClose(_wakeup[1]);
+#endif
     _wakeup[0] = -1;
     _wakeup[1] = -1;
 }
 
-void SocketPoll::startThread()
+bool SocketPoll::startThread()
 {
+    assert(!_runOnClientThread);
+
     if (!_threadStarted)
     {
         _threadStarted = true;
+        _threadFinished = false;
+        _stop = false;
         try
         {
             _thread = std::thread(&SocketPoll::pollingThreadEntry, this);
+            return true;
         }
         catch (const std::exception& exc)
         {
@@ -104,11 +130,16 @@ void SocketPoll::startThread()
             _threadStarted = false;
         }
     }
+
+    return false;
 }
 
 void SocketPoll::joinThread()
 {
-    addCallback([this](){ removeSockets(); });
+    addCallback([this]()
+                {
+                    removeSockets();
+                });
     stop();
     if (_threadStarted && _thread.joinable())
     {
@@ -122,14 +153,47 @@ void SocketPoll::joinThread()
     }
 }
 
+void SocketPoll::pollingThreadEntry()
+{
+    try
+    {
+        Util::setThreadName(_name);
+        LOG_INF("Starting polling thread [" << _name << "].");
+
+        _owner = std::this_thread::get_id();
+        LOG_DBG("Thread affinity of " << _name << " set to " <<
+                Log::to_string(_owner) << ".");
+
+        // Invoke the virtual implementation.
+        pollingThread();
+
+        // Release sockets.
+        _pollSockets.clear();
+        _newSockets.clear();
+    }
+    catch (const std::exception& exc)
+    {
+        LOG_ERR("Exception in polling thread [" << _name << "]: " << exc.what());
+    }
+
+    _threadFinished = true;
+}
+
 void SocketPoll::wakeupWorld()
 {
     for (const auto& fd : getWakeupsArray())
         wakeup(fd);
 }
 
-void SocketPoll::insertNewWebSocketSync(const Poco::URI &uri, const std::shared_ptr<SocketHandlerInterface>& websocketHandler)
+void SocketPoll::insertNewWebSocketSync(
+#ifndef MOBILEAPP
+                                        const Poco::URI &uri,
+#else
+                                        int peerSocket,
+#endif
+                                        const std::shared_ptr<SocketHandlerInterface>& websocketHandler)
 {
+#ifndef MOBILEAPP
     LOG_INF("Connecting to " << uri.getHost() << " : " << uri.getPort() << " : " << uri.getPath());
 
     // FIXME: put this in a ClientSocket class ?
@@ -161,11 +225,10 @@ void SocketPoll::insertNewWebSocketSync(const Poco::URI &uri, const std::shared_
             {
                 int fd = socket(ai->ai_addr->sa_family, SOCK_STREAM | SOCK_NONBLOCK, 0);
                 int res = connect(fd, ai->ai_addr, ai->ai_addrlen);
-                // FIXME: SSL sockets presumably need some setup, checking etc. and ... =)
                 if (fd < 0 || (res < 0 && errno != EINPROGRESS))
                 {
                     LOG_ERR("Failed to connect to " << uri.getHost());
-                    close(fd);
+                    ::close(fd);
                 }
                 else
                 {
@@ -203,7 +266,7 @@ void SocketPoll::insertNewWebSocketSync(const Poco::URI &uri, const std::shared_
                     else
                     {
                         LOG_ERR("Failed to allocate socket for client websocket " << uri.getHost());
-                        close(fd);
+                        ::close(fd);
                     }
 
                     break;
@@ -215,6 +278,32 @@ void SocketPoll::insertNewWebSocketSync(const Poco::URI &uri, const std::shared_
     }
     else
         LOG_ERR("Failed to lookup client websocket host '" << uri.getHost() << "' skipping");
+#else
+    LOG_INF("Connecting to " << peerSocket);
+    int fd = fakeSocketSocket();
+    int res = fakeSocketConnect(fd, peerSocket);
+    if (fd < 0 || (res < 0 && errno != EINPROGRESS))
+    {
+        LOG_ERR("Failed to connect to the 'wsd' socket");
+        fakeSocketClose(fd);
+    }
+    else
+    {
+        std::shared_ptr<StreamSocket> socket;
+        socket = StreamSocket::create<StreamSocket>(fd, true, websocketHandler);
+        if (socket)
+        {
+            LOG_TRC("Sending 'hello' instead of HTTP GET for now");
+            socket->send("hello");
+            insertNewSocket(socket);
+        }
+        else
+        {
+            LOG_ERR("Failed to allocate socket for client websocket");
+            fakeSocketClose(fd);
+        }
+    }
+#endif
 }
 
 void ServerSocket::dumpState(std::ostream& os)
@@ -222,7 +311,6 @@ void ServerSocket::dumpState(std::ostream& os)
     os << "\t" << getFD() << "\t<accept>\n";
 }
 
-
 void SocketDisposition::execute()
 {
     // We should have hard ownership of this socket.
@@ -230,7 +318,7 @@ void SocketDisposition::execute()
     if (_socketMove)
     {
         // Drop pretentions of ownership before _socketMove.
-        _socket->setThreadOwner(std::thread::id(0));
+        _socket->setThreadOwner(std::thread::id());
         _socketMove(_socket);
     }
     _socketMove = nullptr;
@@ -259,7 +347,7 @@ void StreamSocket::dumpState(std::ostream& os)
     if (_inBuffer.size() > 0)
         Util::dumpHex(os, "\t\tinBuffer:\n", "\t\t", _inBuffer);
     if (_outBuffer.size() > 0)
-        Util::dumpHex(os, "\t\toutBuffer:\n", "\t\t", _inBuffer);
+        Util::dumpHex(os, "\t\toutBuffer:\n", "\t\t", _outBuffer);
 }
 
 void StreamSocket::send(Poco::Net::HTTPResponse& response)
@@ -288,6 +376,7 @@ void SocketPoll::dumpState(std::ostream& os)
 /// Returns true on success only.
 bool ServerSocket::bind(Type type, int port)
 {
+#ifndef MOBILEAPP
     // Enable address reuse to avoid stalling after
     // recycling, when previous socket is TIME_WAIT.
     //TODO: Might be worth refactoring out.
@@ -332,8 +421,13 @@ bool ServerSocket::bind(Type type, int port)
         LOG_SYS("Failed to bind to: " << (_type == Socket::Type::IPv4 ? "IPv4" : "IPv6") << " port: " << port);
 
     return rc == 0;
+#else
+    return true;
+#endif
 }
 
+#ifndef MOBILEAPP
+
 bool StreamSocket::parseHeader(const char *clientName,
                                Poco::MemoryInputStream &message,
                                Poco::Net::HTTPRequest &request,
@@ -375,7 +469,7 @@ bool StreamSocket::parseHeader(const char *clientName,
                 logger << " / " << it.first << ": " << it.second;
             }
 
-            LOG_END(logger);
+            LOG_END(logger, true);
         }
 
         const std::streamsize contentLength = request.getContentLength();
@@ -406,7 +500,6 @@ bool StreamSocket::parseHeader(const char *clientName,
     return true;
 }
 
-
 namespace HttpHelper
 {
     void sendUncompressedFileContent(const std::shared_ptr<StreamSocket>& socket,
@@ -514,4 +607,18 @@ namespace HttpHelper
     }
 }
 
+bool StreamSocket::sniffSSL() const
+{
+    // Only sniffing the first bytes of a sockte.
+    if (_bytesSent > 0 || _bytesRecvd != _inBuffer.size() || _bytesRecvd < 6)
+        return false;
+
+    // 0x0000  16 03 01 02 00 01 00 01
+    return (_inBuffer[0] == 0x16 && // HANDSHAKE
+            _inBuffer[1] == 0x03 && // SSL 3.0 / TLS 1.x
+            _inBuffer[5] == 0x01);  // Handshake: CLIENT_HELLO
+}
+
+#endif // !MOBILEAPP
+
 /* vim:set shiftwidth=4 softtabstop=4 expandtab: */
