@@ -26,7 +26,7 @@
 
 class WebSocketHandler : public SocketHandlerInterface
 {
-protected:
+private:
     /// The socket that owns us (we can't own it).
     std::weak_ptr<StreamSocket> _socket;
 
@@ -36,7 +36,9 @@ protected:
     std::vector<char> _wsPayload;
     std::atomic<bool> _shuttingDown;
     bool _isClient;
+    bool _isMasking;
 
+protected:
     struct WSFrameMask
     {
         static const unsigned char Fin = 0x80;
@@ -48,11 +50,12 @@ protected:
 
 public:
     /// Perform upgrade ourselves, or select a client web socket.
-    WebSocketHandler(bool isClient = false) :
+    WebSocketHandler(bool isClient = false, bool isMasking = true) :
         _lastPingSentTime(std::chrono::steady_clock::now()),
         _pingTimeUs(0),
         _shuttingDown(false),
-        _isClient(isClient)
+        _isClient(isClient),
+        _isMasking(isClient && isMasking)
     {
     }
 
@@ -65,7 +68,8 @@ public:
                   std::chrono::milliseconds(InitialPingDelayMs)),
         _pingTimeUs(0),
         _shuttingDown(false),
-        _isClient(false)
+        _isClient(false),
+        _isMasking(false)
     {
         upgradeToWebSocket(request);
     }
@@ -74,7 +78,7 @@ public:
     void onConnect(const std::shared_ptr<StreamSocket>& socket) override
     {
         _socket = socket;
-        LOG_TRC("#" << socket->getFD() << " Connected to WS Handler 0x" << std::hex << this << std::dec);
+        LOG_TRC("#" << socket->getFD() << " Connected to WS Handler " << this);
     }
 
     /// Status codes sent to peer on shutdown.
@@ -101,7 +105,7 @@ public:
         std::shared_ptr<StreamSocket> socket = _socket.lock();
         if (socket == nullptr)
         {
-            LOG_ERR("No socket associated with WebSocketHandler 0x" << std::hex << this << std::dec);
+            LOG_ERR("No socket associated with WebSocketHandler " << this);
             return;
         }
 
@@ -109,6 +113,7 @@ public:
                 static_cast<unsigned>(statusCode) << ", message: " << statusMessage);
         _shuttingDown = true;
 
+#ifndef MOBILEAPP
         const size_t len = statusMessage.size();
         std::vector<char> buf(2 + len);
         buf[0] = ((((int)statusCode) >> 8) & 0xff);
@@ -118,6 +123,7 @@ public:
                                   | static_cast<char>(WSOpCode::Close);
 
         sendFrame(socket, buf.data(), buf.size(), flags);
+#endif
     }
 
     bool handleOneIncomingMessage(const std::shared_ptr<StreamSocket>& socket)
@@ -125,17 +131,19 @@ public:
         assert(socket && "Expected a valid socket instance.");
 
         // websocket fun !
-        const size_t len = socket->_inBuffer.size();
+        const size_t len = socket->getInBuffer().size();
 
         if (len == 0)
             return false; // avoid logging.
 
-        LOG_TRC("#" << socket->getFD() << ": Incoming WebSocket data of " << len << " bytes: " << std::string(socket->_inBuffer.data(), socket->_inBuffer.size()));
-
+#ifndef MOBILEAPP
         if (len < 2) // partial read
+        {
+            LOG_TRC("#" << socket->getFD() << ": Still incomplete WebSocket message, have " << len << " bytes");
             return false;
+        }
 
-        unsigned char *p = reinterpret_cast<unsigned char*>(&socket->_inBuffer[0]);
+        unsigned char *p = reinterpret_cast<unsigned char*>(&socket->getInBuffer()[0]);
         const bool fin = p[0] & 0x80;
         const WSOpCode code = static_cast<WSOpCode>(p[0] & 0x0f);
         const bool hasMask = p[1] & 0x80;
@@ -146,7 +154,10 @@ public:
         if (payloadLen == 126) // 2 byte length
         {
             if (len < 2 + 2)
+            {
+                LOG_TRC("#" << socket->getFD() << ": Still incomplete WebSocket message, have " << len << " bytes");
                 return false;
+            }
 
             payloadLen = (((unsigned)p[2]) << 8) | ((unsigned)p[3]);
             headerLen += 2;
@@ -154,8 +165,10 @@ public:
         else if (payloadLen == 127) // 8 byte length
         {
             if (len < 2 + 8)
+            {
+                LOG_TRC("#" << socket->getFD() << ": Still incomplete WebSocket message, have " << len << " bytes");
                 return false;
-
+            }
             payloadLen = ((((uint64_t)p[9]) <<  0) + (((uint64_t)p[8]) <<  8) +
                           (((uint64_t)p[7]) << 16) + (((uint64_t)p[6]) << 24) +
                           (((uint64_t)p[5]) << 32) + (((uint64_t)p[4]) << 40) +
@@ -174,9 +187,12 @@ public:
 
         if (payloadLen + headerLen > len)
         { // partial read wait for more data.
+            LOG_TRC("#" << socket->getFD() << ": Still incomplete WebSocket message, have " << len << " bytes, message is " << payloadLen + headerLen << " bytes");
             return false;
         }
 
+        LOG_TRC("#" << socket->getFD() << ": Incoming WebSocket data of " << len << " bytes: " << Util::stringifyHexLine(socket->getInBuffer(), 0, std::min((size_t)32, len)));
+
         data = p + headerLen;
 
         if (hasMask)
@@ -188,15 +204,23 @@ public:
                 *wsData++ = data[i] ^ mask[i % 4];
         } else
             _wsPayload.insert(_wsPayload.end(), data, data + payloadLen);
+#else
+        unsigned char * const p = reinterpret_cast<unsigned char*>(&socket->getInBuffer()[0]);
+        _wsPayload.insert(_wsPayload.end(), p, p + len);
+        const size_t headerLen = 0;
+        const size_t payloadLen = len;
+#endif
 
         assert(_wsPayload.size() >= payloadLen);
 
-        socket->_inBuffer.erase(socket->_inBuffer.begin(), socket->_inBuffer.begin() + headerLen + payloadLen);
+        socket->getInBuffer().erase(socket->getInBuffer().begin(), socket->getInBuffer().begin() + headerLen + payloadLen);
+
+#ifndef MOBILEAPP
 
         // FIXME: fin, aggregating payloads into _wsPayload etc.
         LOG_TRC("#" << socket->getFD() << ": Incoming WebSocket message code " << static_cast<unsigned>(code) <<
                 ", fin? " << fin << ", mask? " << hasMask << ", payload length: " << _wsPayload.size() <<
-                ", residual socket data: " << socket->_inBuffer.size() << " bytes.");
+                ", residual socket data: " << socket->getInBuffer().size() << " bytes.");
 
         bool doClose = false;
 
@@ -241,6 +265,12 @@ public:
             break;
         }
 
+#else
+        handleMessage(true, WSOpCode::Binary, _wsPayload);
+
+#endif
+
+#ifndef MOBILEAPP
         if (doClose)
         {
             if (!_shuttingDown)
@@ -268,6 +298,7 @@ public:
             // TCP Close.
             socket->closeConnection();
         }
+#endif
 
         _wsPayload.clear();
 
@@ -277,17 +308,28 @@ public:
     /// Implementation of the SocketHandlerInterface.
     virtual void handleIncomingMessage(SocketDisposition&) override
     {
+        // LOG_TRC("***** WebSocketHandler::handleIncomingMessage()");
+
         std::shared_ptr<StreamSocket> socket = _socket.lock();
+
+#ifdef MOBILEAPP
+        // No separate "upgrade" is going on
+        if (socket != nullptr && !socket->isWebSocket())
+            socket->setWebSocket();
+#endif
+
         if (socket == nullptr)
         {
-            LOG_ERR("No socket associated with WebSocketHandler 0x" << std::hex << this << std::dec);
+            LOG_ERR("No socket associated with WebSocketHandler " << this);
         }
+#ifndef MOBILEAPP
         else if (_isClient && !socket->isWebSocket())
             handleClientUpgrade();
+#endif
         else
         {
             while (handleOneIncomingMessage(socket))
-                ; // can have multiple msgs in one recv'd packet.
+                ; // might have multiple messages in the accumulated buffer.
         }
     }
 
@@ -303,6 +345,7 @@ public:
         return POLLIN;
     }
 
+#ifndef MOBILEAPP
     /// Send a ping message
     void sendPingOrPong(std::chrono::steady_clock::time_point now,
                         const char* data, const size_t len,
@@ -340,10 +383,12 @@ public:
         assert(_isClient);
         sendPingOrPong(now, data, len, WSOpCode::Pong, socket);
     }
+#endif
 
     /// Do we need to handle a timeout ?
     void checkTimeout(std::chrono::steady_clock::time_point now) override
     {
+#ifndef MOBILEAPP
         if (_isClient)
             return;
 
@@ -355,6 +400,7 @@ public:
             if (socket)
                 sendPing(now, socket);
         }
+#endif
     }
 
     /// By default rely on the socket buffer.
@@ -381,14 +427,14 @@ public:
         return sendFrame(socket, data, len, WSFrameMask::Fin | static_cast<unsigned char>(code), flush);
     }
 
-protected:
+private:
 
     /// Sends a WebSocket frame given the data, length, and flags.
     /// Returns the number of bytes written (including frame overhead) on success,
     /// 0 for closed/invalid socket, and -1 for other errors.
-    static int sendFrame(const std::shared_ptr<StreamSocket>& socket,
-                         const char* data, const size_t len,
-                         const unsigned char flags, const bool flush = true)
+    int sendFrame(const std::shared_ptr<StreamSocket>& socket,
+                  const char* data, const size_t len,
+                  unsigned char flags, const bool flush = true) const
     {
         if (!socket || data == nullptr || len == 0)
             return -1;
@@ -397,24 +443,27 @@ protected:
             return 0;
 
         socket->assertCorrectThread();
-        std::vector<char>& out = socket->_outBuffer;
+        std::vector<char>& out = socket->getOutBuffer();
+
+#ifndef MOBILEAPP
         const size_t oldSize = out.size();
 
         out.push_back(flags);
 
+        int maskFlag = _isMasking ? 0x80 : 0;
         if (len < 126)
         {
-            out.push_back((char)len);
+            out.push_back((char)(len | maskFlag));
         }
         else if (len <= 0xffff)
         {
-            out.push_back((char)126);
+            out.push_back((char)(126 | maskFlag));
             out.push_back(static_cast<char>((len >> 8) & 0xff));
             out.push_back(static_cast<char>((len >> 0) & 0xff));
         }
         else
         {
-            out.push_back((char)127);
+            out.push_back((char)(127 | maskFlag));
             out.push_back(static_cast<char>((len >> 56) & 0xff));
             out.push_back(static_cast<char>((len >> 48) & 0xff));
             out.push_back(static_cast<char>((len >> 40) & 0xff));
@@ -425,21 +474,59 @@ protected:
             out.push_back(static_cast<char>((len >> 0) & 0xff));
         }
 
-        // Copy the data.
-        out.insert(out.end(), data, data + len);
-        const size_t size = out.size() - oldSize;
+        if (_isMasking)
+        { // flip some top bits - perhaps it helps.
+            size_t mask = out.size();
 
+            out.push_back(static_cast<char>(0x81));
+            out.push_back(static_cast<char>(0x76));
+            out.push_back(static_cast<char>(0x81));
+            out.push_back(static_cast<char>(0x76));
+
+            // Copy the data.
+            out.insert(out.end(), data, data + len);
+
+            // Mask it.
+            for (size_t i = 4; i < out.size() - mask; ++i)
+                out[mask + i] = out[mask + i] ^ out[mask + (i%4)];
+        }
+        else
+        {
+            // Copy the data.
+            out.insert(out.end(), data, data + len);
+        }
+        const size_t size = out.size() - oldSize;
+#else
+        LOG_TRC("WebSocketHandle::sendFrame: Writing to #" << socket->getFD() << " " << len << " bytes");
+        assert(flush);
+        assert(out.size() == 0);
+
+        out.insert(out.end(), data, data + len);
+        const size_t size = out.size();
+#endif
         if (flush)
             socket->writeOutgoingData();
 
         return size;
     }
 
+protected:
+
     /// To be overriden to handle the websocket messages the way you need.
     virtual void handleMessage(bool /*fin*/, WSOpCode /*code*/, std::vector<char> &/*data*/)
     {
     }
 
+    std::weak_ptr<StreamSocket>& getSocket()
+    {
+        return _socket;
+    }
+
+    void setSocket(const std::weak_ptr<StreamSocket>& socket)
+    {
+        _socket = socket;
+    }
+
     void dumpState(std::ostream& os) override;
 
 private:
@@ -464,6 +551,7 @@ protected:
         LOG_TRC("#" << socket->getFD() << ": Upgrading to WebSocket.");
         assert(!socket->isWebSocket());
 
+#ifndef MOBILEAPP
         // create our websocket goodness ...
         const int wsVersion = std::stoi(req.get("Sec-WebSocket-Version", "13"));
         const std::string wsKey = req.get("Sec-WebSocket-Key", "");
@@ -487,34 +575,36 @@ protected:
         const std::string res = oss.str();
         LOG_TRC("#" << socket->getFD() << ": Sending WS Upgrade response: " << res);
         socket->send(res);
+#endif
         setWebSocket();
     }
 
+#ifndef MOBILEAPP
     // Handle incoming upgrade to full socket as client WS.
     void handleClientUpgrade()
     {
         std::shared_ptr<StreamSocket> socket = _socket.lock();
 
-        LOG_TRC("Incoming client websocket upgrade response: " << std::string(&socket->_inBuffer[0], socket->_inBuffer.size()));
+        LOG_TRC("Incoming client websocket upgrade response: " << std::string(&socket->getInBuffer()[0], socket->getInBuffer().size()));
 
         bool bOk = false;
         size_t responseSize = 0;
 
         try
         {
-            Poco::MemoryInputStream message(&socket->_inBuffer[0], socket->_inBuffer.size());;
+            Poco::MemoryInputStream message(&socket->getInBuffer()[0], socket->getInBuffer().size());;
             Poco::Net::HTTPResponse response;
 
             response.read(message);
 
             {
                 static const std::string marker("\r\n\r\n");
-                auto itBody = std::search(socket->_inBuffer.begin(),
-                                          socket->_inBuffer.end(),
+                auto itBody = std::search(socket->getInBuffer().begin(),
+                                          socket->getInBuffer().end(),
                                           marker.begin(), marker.end());
 
-                if (itBody != socket->_inBuffer.end())
-                    responseSize = itBody - socket->_inBuffer.begin() + marker.size();
+                if (itBody != socket->getInBuffer().end())
+                    responseSize = itBody - socket->getInBuffer().begin() + marker.size();
             }
 
             if (response.getStatus() == Poco::Net::HTTPResponse::HTTP_SWITCHING_PROTOCOLS &&
@@ -554,6 +644,7 @@ protected:
         setWebSocket();
         socket->eraseFirstInputBytes(responseSize);
     }
+#endif
 
     void setWebSocket()
     {
