@@ -296,7 +296,7 @@ def writeTranslations(onlineDir, translationsDir, strings):
                         if text == entry.msgid:
                             translations[entry.msgid] = entry.msgstr
 
-        f = open(onlineDir + '/loleaflet/dist/l10n/uno/' + lang + '.json', 'w')
+        f = open(onlineDir + '/loleaflet/l10n/uno/' + lang + '.json', 'w')
         f.write('{\n')
 
         writeComma = False
@@ -305,7 +305,7 @@ def writeTranslations(onlineDir, translationsDir, strings):
                 f.write(',\n')
             else:
                 writeComma = True
-            f.write(('"' + key + '":"' + translations[key] + '"').encode('utf-8'))
+            f.write(('"' + key.replace('"','\\\"') + '":"' + translations[key].replace('"','\\\"') + '"').encode('utf-8'))
 
         f.write('\n}\n')
 
