@@ -43,8 +43,9 @@ public:
 
     bool filterSendMessage(const char* data, const size_t len, const WSOpCode /* code */, const bool /* flush */, int& /*unitReturn*/) override
     {
-        std::string message(data, len);
-        if (message == "saveas: url=" + helpers::getTestServerURI() + "/something%20wopi/files/1?access_token=anything filename=hello%20world.pdf")
+        const std::string message(data, len);
+        const std::string expected("saveas: url=" + helpers::getTestServerURI() + "/something%20wopi/files/1?access_token=anything filename=hello%20world.pdf");
+        if (message.find(expected) == 0)
         {
             // successfully exit the test if we also got the outgoing message
             // notifying about saving the file
