@@ -318,7 +318,7 @@ std::string getResponseString(T& ws, const std::string& prefix, const std::strin
 }
 
 template <typename T>
-std::string assertResponseString(T& ws, const std::string& prefix, const std::string testname)
+std::string assertResponseString(T& ws, const std::string& prefix, const std::string& testname)
 {
     const auto res = getResponseString(ws, prefix, testname);
     CPPUNIT_ASSERT_EQUAL(prefix, res.substr(0, prefix.length()));
@@ -327,7 +327,7 @@ std::string assertResponseString(T& ws, const std::string& prefix, const std::st
 
 /// Assert that we don't get a response with the given prefix.
 template <typename T>
-std::string assertNotInResponse(T& ws, const std::string& prefix, const std::string testname)
+std::string assertNotInResponse(T& ws, const std::string& prefix, const std::string& testname)
 {
     const auto res = getResponseString(ws, prefix, testname, 1000);
     CPPUNIT_ASSERT_MESSAGE(testname + "Did not expect getting message [" + res + "].", res.empty());
@@ -378,7 +378,7 @@ connectLOKit(const Poco::URI& uri,
         catch (const std::exception& ex)
         {
             TST_LOG_END;
-            TST_LOG("Error connecting: " << ex.what());
+            TST_LOG("Connection problem: " << ex.what());
         }
 
         std::this_thread::sleep_for(std::chrono::milliseconds(POLL_TIMEOUT_MS));
