@@ -12,13 +12,17 @@
 #include "ClientSession.hpp"
 
 #include <fstream>
+#include <sstream>
+#include <memory>
 
 #include <Poco/Net/HTTPResponse.h>
+#include <Poco/StreamCopier.h>
 #include <Poco/StringTokenizer.h>
 #include <Poco/URI.h>
 
 #include "DocumentBroker.hpp"
 #include "LOOLWSD.hpp"
+#include "Storage.hpp"
 #include <common/Common.hpp>
 #include <common/Log.hpp>
 #include <common/Protocol.hpp>
@@ -41,7 +45,14 @@ ClientSession::ClientSession(const std::string& id,
     _isDocumentOwner(false),
     _isAttached(false),
     _isViewLoaded(false),
-    _keyEvents(1)
+    _keyEvents(1),
+    _clientVisibleArea(0, 0, 0, 0),
+    _clientSelectedPart(-1),
+    _tileWidthPixel(0),
+    _tileHeightPixel(0),
+    _tileWidthTwips(0),
+    _tileHeightTwips(0),
+    _isTextDocument(false)
 {
     const size_t curConnections = ++LOOLWSD::NumConnections;
     LOG_INF("ClientSession ctor [" << getName() << "], current number of connections: " << curConnections);
@@ -55,6 +66,7 @@ ClientSession::~ClientSession()
 
 void ClientSession::handleIncomingMessage(SocketDisposition &disposition)
 {
+    // LOG_TRC("***** ClientSession::handleIncomingMessage()");
     if (UnitWSD::get().filterHandleRequest(
             UnitWSD::TestRequest::Client, disposition, *this))
         return;
@@ -112,10 +124,9 @@ bool ClientSession::_handleInput(const char *buffer, int length)
 
         return true;
     }
-
-    if (tokens[0] == "load")
+    else if (tokens[0] == "load")
     {
-        if (_docURL != "")
+        if (getDocURL() != "")
         {
             sendTextFrame("error: cmd=load kind=docalreadyloaded");
             return false;
@@ -124,6 +135,7 @@ bool ClientSession::_handleInput(const char *buffer, int length)
         return loadDocument(buffer, length, tokens, docBroker);
     }
     else if (tokens[0] != "canceltiles" &&
+             tokens[0] != "tileprocessed" &&
              tokens[0] != "clientzoom" &&
              tokens[0] != "clientvisiblearea" &&
              tokens[0] != "outlinestate" &&
@@ -159,12 +171,14 @@ bool ClientSession::_handleInput(const char *buffer, int length)
              tokens[0] != "useractive" &&
              tokens[0] != "userinactive" &&
              tokens[0] != "paintwindow" &&
-             tokens[0] != "windowcommand")
+             tokens[0] != "windowcommand" &&
+             tokens[0] != "signdocument" &&
+             tokens[0] != "asksignaturestatus")
     {
         sendTextFrame("error: cmd=" + tokens[0] + " kind=unknown");
         return false;
     }
-    else if (_docURL == "")
+    else if (getDocURL() == "")
     {
         sendTextFrame("error: cmd=" + tokens[0] + " kind=nodocloaded");
         return false;
@@ -258,6 +272,93 @@ bool ClientSession::_handleInput(const char *buffer, int length)
             docBroker->broadcastMessage("commandresult: { \"command\": \"savetostorage\", \"success\": true }");
         }
     }
+    else if (tokens[0] == "clientvisiblearea")
+    {
+        int x;
+        int y;
+        int width;
+        int height;
+        if (tokens.size() != 5 ||
+            !getTokenInteger(tokens[1], "x", x) ||
+            !getTokenInteger(tokens[2], "y", y) ||
+            !getTokenInteger(tokens[3], "width", width) ||
+            !getTokenInteger(tokens[4], "height", height))
+        {
+            sendTextFrame("error: cmd=clientvisiblearea kind=syntax");
+            return false;
+        }
+        else
+        {
+            _clientVisibleArea = Util::Rectangle(x, y, width, height);
+            resetWireIdMap();
+            return forwardToChild(std::string(buffer, length), docBroker);
+        }
+    }
+    else if (tokens[0] == "setclientpart")
+    {
+        if(!_isTextDocument)
+        {
+            int temp;
+            if (tokens.size() != 2 ||
+                !getTokenInteger(tokens[1], "part", temp))
+            {
+                sendTextFrame("error: cmd=setclientpart kind=syntax");
+                return false;
+            }
+            else
+            {
+                _clientSelectedPart = temp;
+                resetWireIdMap();
+                return forwardToChild(std::string(buffer, length), docBroker);
+            }
+        }
+    }
+    else if (tokens[0] == "clientzoom")
+    {
+        int tilePixelWidth, tilePixelHeight, tileTwipWidth, tileTwipHeight;
+        if (tokens.size() != 5 ||
+            !getTokenInteger(tokens[1], "tilepixelwidth", tilePixelWidth) ||
+            !getTokenInteger(tokens[2], "tilepixelheight", tilePixelHeight) ||
+            !getTokenInteger(tokens[3], "tiletwipwidth", tileTwipWidth) ||
+            !getTokenInteger(tokens[4], "tiletwipheight", tileTwipHeight))
+        {
+            sendTextFrame("error: cmd=clientzoom kind=syntax");
+            return false;
+        }
+        else
+        {
+            _tileWidthPixel = tilePixelWidth;
+            _tileHeightPixel = tilePixelHeight;
+            _tileWidthTwips = tileTwipWidth;
+            _tileHeightTwips = tileTwipHeight;
+            resetWireIdMap();
+            return forwardToChild(std::string(buffer, length), docBroker);
+        }
+    }
+    else if (tokens[0] == "tileprocessed")
+    {
+        std::string tileID;
+        if (tokens.size() != 2 ||
+            !getTokenString(tokens[1], "tile", tileID))
+        {
+            sendTextFrame("error: cmd=tileprocessed kind=syntax");
+            return false;
+        }
+
+        auto iter = std::find_if(_tilesOnFly.begin(), _tilesOnFly.end(),
+        [&tileID](const std::pair<std::string, std::chrono::steady_clock::time_point>& curTile)
+        {
+            return curTile.first == tileID;
+        });
+
+        if(iter != _tilesOnFly.end())
+            _tilesOnFly.erase(iter);
+        else
+            LOG_INF("Tileprocessed message with an unknown tile ID");
+
+        docBroker->sendRequestedTiles(shared_from_this());
+        return true;
+    }
     else
     {
         if (tokens[0] == "key")
@@ -302,24 +403,26 @@ bool ClientSession::loadDocument(const char* /*buffer*/, int /*length*/,
 
         std::ostringstream oss;
         oss << "load";
-        oss << " url=" << docBroker->getPublicUri().toString();
+        oss << " url=" << docBroker->getPublicUri().toString();;
 
-        if (!_userId.empty() && !_userName.empty())
+        if (!getUserId().empty() && !getUserName().empty())
         {
             std::string encodedUserId;
-            Poco::URI::encode(_userId, "", encodedUserId);
+            Poco::URI::encode(getUserId(), "", encodedUserId);
             oss << " authorid=" << encodedUserId;
+            oss << " xauthorid=" << LOOLWSD::anonymizeUsername(getUserId());
 
             std::string encodedUserName;
-            Poco::URI::encode(_userName, "", encodedUserName);
+            Poco::URI::encode(getUserName(), "", encodedUserName);
             oss << " author=" << encodedUserName;
+            oss << " xauthor=" << LOOLWSD::anonymizeUsername(getUserName());
         }
 
-        if (!_userExtraInfo.empty())
+        if (!getUserExtraInfo().empty())
         {
             std::string encodedUserExtraInfo;
-            Poco::URI::encode(_userExtraInfo, "", encodedUserExtraInfo);
-            oss << " authorextrainfo=" << encodedUserExtraInfo;
+            Poco::URI::encode(getUserExtraInfo(), "", encodedUserExtraInfo);
+            oss << " authorextrainfo=" << encodedUserExtraInfo; //TODO: could this include PII?
         }
 
         oss << " readonly=" << isReadOnly();
@@ -329,26 +432,26 @@ bool ClientSession::loadDocument(const char* /*buffer*/, int /*length*/,
             oss << " part=" << loadPart;
         }
 
-        if (_haveDocPassword)
+        if (getHaveDocPassword())
         {
-            oss << " password=" << _docPassword;
+            oss << " password=" << getDocPassword();
         }
 
-        if (!_lang.empty())
+        if (!getLang().empty())
         {
-            oss << " lang=" << _lang;
+            oss << " lang=" << getLang();
         }
 
-        if (!_watermarkText.empty())
+        if (!getWatermarkText().empty())
         {
             std::string encodedWatermarkText;
-            Poco::URI::encode(_watermarkText, "", encodedWatermarkText);
+            Poco::URI::encode(getWatermarkText(), "", encodedWatermarkText);
             oss << " watermarkText=" << encodedWatermarkText;
         }
 
-        if (!_docOptions.empty())
+        if (!getDocOptions().empty())
         {
-            oss << " options=" << _docOptions;
+            oss << " options=" << getDocOptions();
         }
 
         return forwardToChild(oss.str(), docBroker);
@@ -477,11 +580,11 @@ bool ClientSession::filterMessage(const std::string& message) const
         }
         else
         {
-                allowed = false;
-                LOG_WRN("No value of id in downloadas message");
+            allowed = false;
+            LOG_WRN("No value of id in downloadas message");
         }
     }
-    else if (tokens[0] == "gettextselection")
+    else if (tokens[0] == "gettextselection" || tokens[0] == ".uno:Copy")
     {
         if (_wopiFileInfo && _wopiFileInfo->_disableCopy)
         {
@@ -528,7 +631,7 @@ int ClientSession::getPollEvents(std::chrono::steady_clock::time_point /* now */
 
 void ClientSession::performWrites()
 {
-    LOG_DBG(getName() << " ClientSession: performing writes");
+    LOG_TRC(getName() << " ClientSession: performing writes.");
 
     std::shared_ptr<Message> item;
     if (_senderQueue.dequeue(item))
@@ -552,7 +655,7 @@ void ClientSession::performWrites()
         }
     }
 
-    LOG_DBG(getName() << " ClientSession: performed write");
+    LOG_TRC(getName() << " ClientSession: performed write.");
 }
 
 bool ClientSession::handleKitToClientMessage(const char* buffer, const int length)
@@ -569,7 +672,9 @@ bool ClientSession::handleKitToClientMessage(const char* buffer, const int lengt
         return false;
     }
 
+#ifndef MOBILEAPP
     LOOLWSD::dumpOutgoingTrace(docBroker->getJailId(), getId(), firstLine);
+#endif
 
     const auto& tokens = payload->tokens();
     if (tokens[0] == "unocommandresult:")
@@ -618,15 +723,19 @@ bool ClientSession::handleKitToClientMessage(const char* buffer, const int lengt
         {
             if (errorCommand == "load")
             {
+                LOG_WRN("Document load failed: " << errorKind);
                 if (errorKind == "passwordrequired:to-view" ||
                     errorKind == "passwordrequired:to-modify" ||
                     errorKind == "wrongpassword")
                 {
                     forwardToClient(payload);
-                    LOG_WRN("Document load failed: " << errorKind);
                     return false;
                 }
             }
+            else
+            {
+                LOG_WRN("Other than load failure: " << errorKind);
+            }
         }
     }
     else if (tokens[0] == "curpart:" && tokens.size() == 2)
@@ -635,6 +744,26 @@ bool ClientSession::handleKitToClientMessage(const char* buffer, const int lengt
         int curPart;
         return getTokenInteger(tokens[1], "part", curPart);
     }
+    else if (tokens[0] == "setpart:" && tokens.size() == 2)
+    {
+        if(!_isTextDocument)
+        {
+            int setPart;
+            if(getTokenInteger(tokens[1], "part", setPart))
+            {
+                _clientSelectedPart = setPart;
+                resetWireIdMap();
+            }
+            else if (stringToInteger(tokens[1], setPart))
+            {
+                _clientSelectedPart = setPart;
+                resetWireIdMap();
+            }
+            else
+                return false;
+         }
+    }
+#ifndef MOBILEAPP
     else if (tokens.size() == 3 && tokens[0] == "saveas:")
     {
         bool isConvertTo = static_cast<bool>(_saveAsSocket);
@@ -727,10 +856,16 @@ bool ClientSession::handleKitToClientMessage(const char* buffer, const int lengt
 
             // Now terminate.
             docBroker->stop("Finished saveas handler.");
+
+            // Remove file and directory
+            Poco::Path path = docBroker->getDocKey();
+            Poco::File(path).remove();
+            Poco::File(path.makeParent()).remove();
         }
 
         return true;
     }
+#endif
     else if (tokens.size() == 2 && tokens[0] == "statechanged:")
     {
         StringTokenizer stateTokens(tokens[1], "=", StringTokenizer::TOK_IGNORE_EMPTY | StringTokenizer::TOK_TRIM);
@@ -738,9 +873,43 @@ bool ClientSession::handleKitToClientMessage(const char* buffer, const int lengt
         {
             docBroker->setModified(stateTokens[1] == "true");
         }
+        else
+        {
+            // Set the initial settings per the user's request.
+            const std::pair<std::string, std::string> unoStatePair = Util::split(tokens[1], '=');
+
+            if (!docBroker->isInitialSettingSet(unoStatePair.first))
+            {
+                docBroker->setInitialSetting(unoStatePair.first);
+                if (unoStatePair.first == ".uno:TrackChanges")
+                {
+                    if ((unoStatePair.second == "true" &&
+                         _wopiFileInfo && _wopiFileInfo->_disableChangeTrackingRecord == WopiStorage::WOPIFileInfo::TriState::True) ||
+                        (unoStatePair.second == "false" &&
+                         _wopiFileInfo && _wopiFileInfo->_disableChangeTrackingRecord == WopiStorage::WOPIFileInfo::TriState::False))
+                    {
+                        // Toggle the TrackChanges state.
+                        LOG_DBG("Forcing " << unoStatePair.first << " toggle per user settings.");
+                        forwardToChild("uno .uno:TrackChanges", docBroker);
+                    }
+                }
+                else if (unoStatePair.first == ".uno:ShowTrackedChanges")
+                {
+                    if ((unoStatePair.second == "true" &&
+                         _wopiFileInfo && _wopiFileInfo->_disableChangeTrackingShow == WopiStorage::WOPIFileInfo::TriState::True) ||
+                        (unoStatePair.second == "false" &&
+                         _wopiFileInfo && _wopiFileInfo->_disableChangeTrackingShow == WopiStorage::WOPIFileInfo::TriState::False))
+                    {
+                        // Toggle the ShowTrackChanges state.
+                        LOG_DBG("Forcing " << unoStatePair.first << " toggle per user settings.");
+                        forwardToChild("uno .uno:ShowTrackedChanges", docBroker);
+                    }
+                }
+            }
+        }
     }
 
-    if (!_isDocPasswordProtected)
+    if (!isDocPasswordProtected())
     {
         if (tokens[0] == "tile:")
         {
@@ -751,6 +920,24 @@ bool ClientSession::handleKitToClientMessage(const char* buffer, const int lengt
             setViewLoaded();
             docBroker->setLoaded();
 
+            for(auto &token : tokens)
+            {
+                // Need to get the initial part id from status message
+                int part = -1;
+                if(getTokenInteger(token, "current", part))
+                {
+                    _clientSelectedPart = part;
+                    resetWireIdMap();
+                }
+
+                // Get document type too
+                std::string docType;
+                if(getTokenString(token, "type", docType))
+                {
+                    _isTextDocument = docType.find("text") != std::string::npos;
+                }
+            }
+
             // Forward the status response to the client.
             return forwardToClient(payload);
         }
@@ -776,7 +963,12 @@ bool ClientSession::handleKitToClientMessage(const char* buffer, const int lengt
         else if (tokens[0] == "invalidatetiles:")
         {
             assert(firstLine.size() == static_cast<std::string::size_type>(length));
-            docBroker->invalidateTiles(firstLine);
+
+            // First forward invalidation
+            bool ret = forwardToClient(payload);
+
+            handleTileInvalidation(firstLine, docBroker);
+            return ret;
         }
         else if (tokens[0] == "invalidatecursor:")
         {
@@ -834,19 +1026,49 @@ bool ClientSession::handleKitToClientMessage(const char* buffer, const int lengt
 
 bool ClientSession::forwardToClient(const std::shared_ptr<Message>& payload)
 {
-    const auto& message = payload->abbr();
-
     if (isCloseFrame())
     {
-        LOG_TRC(getName() << ": peer began the closing handshake. Dropping forward message [" << message << "].");
+        LOG_TRC(getName() << ": peer began the closing handshake. Dropping forward message [" << payload->abbr() << "].");
         return true;
     }
 
     enqueueSendMessage(payload);
-
     return true;
 }
 
+void ClientSession::enqueueSendMessage(const std::shared_ptr<Message>& data)
+{
+    const std::shared_ptr<DocumentBroker> docBroker = _docBroker.lock();
+    // If in the correct thread - no need for wakeups.
+    if (docBroker)
+        docBroker->assertCorrectThread();
+
+    const std::string command = data->firstToken();
+    std::unique_ptr<TileDesc> tile;
+    if (command == "tile:")
+    {
+        // Avoid sending tile if it has the same wireID as the previously sent tile
+        tile.reset(new TileDesc(TileDesc::parse(data->firstLine())));
+        const std::string tileID = generateTileID(*tile);
+        auto iter = _oldWireIds.find(tileID);
+        if(iter != _oldWireIds.end() && tile->getWireId() != 0 && tile->getWireId() == iter->second)
+        {
+            LOG_INF("WSD filters out a tile with the same wireID: " <<  tile->serialize("tile:"));
+            return;
+        }
+    }
+
+    LOG_TRC(getName() << " enqueueing client message " << data->id());
+    size_t sizeBefore = _senderQueue.size();
+    size_t newSize = _senderQueue.enqueue(data);
+
+    // Track sent tile
+    if (tile)
+    {
+        traceTileBySend(*tile, sizeBefore == newSize);
+    }
+}
+
 Authorization ClientSession::getAuthorization() const
 {
     Poco::URI::QueryParameters queryParams = _uriPublic.getQueryParameters();
@@ -875,6 +1097,56 @@ Authorization ClientSession::getAuthorization() const
     return Authorization();
 }
 
+void ClientSession::addTileOnFly(const TileDesc& tile)
+{
+    _tilesOnFly.push_back({generateTileID(tile), std::chrono::steady_clock::now()});
+}
+
+void ClientSession::clearTilesOnFly()
+{
+    _tilesOnFly.clear();
+}
+
+void ClientSession::removeOutdatedTilesOnFly()
+{
+    // Check only the beginning of the list, tiles are ordered by timestamp
+    bool continueLoop = true;
+    while(!_tilesOnFly.empty() && continueLoop)
+    {
+        auto tileIter = _tilesOnFly.begin();
+        double elapsedTimeMs = std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::steady_clock::now() - tileIter->second).count();
+        if(elapsedTimeMs > TILE_ROUNDTRIP_TIMEOUT_MS)
+        {
+            LOG_WRN("Tracker tileID was dropped because of time out. Tileprocessed message did not arrive");
+            _tilesOnFly.erase(tileIter);
+        }
+        else
+            continueLoop = false;
+    }
+}
+
+size_t ClientSession::countIdenticalTilesOnFly(const TileDesc& tile) const
+{
+    size_t count = 0;
+    std::string tileID = generateTileID(tile);
+    for(auto& tileItem : _tilesOnFly)
+    {
+        if(tileItem.first == tileID)
+            ++count;
+    }
+    return count;
+}
+
+Util::Rectangle ClientSession::getNormalizedVisibleArea() const
+{
+    Util::Rectangle normalizedVisArea;
+    normalizedVisArea.setLeft(std::max(_clientVisibleArea.getLeft(), 0));
+    normalizedVisArea.setTop(std::max(_clientVisibleArea.getTop(), 0));
+    normalizedVisArea.setRight(_clientVisibleArea.getRight());
+    normalizedVisArea.setBottom(_clientVisibleArea.getBottom());
+    return normalizedVisArea;
+}
+
 void ClientSession::onDisconnect()
 {
     LOG_INF(getName() << " Disconnected, current number of connections: " << LOOLWSD::NumConnections);
@@ -941,7 +1213,7 @@ void ClientSession::dumpState(std::ostream& os)
        << "\n\t\tisAttached: " << _isAttached
        << "\n\t\tkeyEvents: " << _keyEvents;
 
-    std::shared_ptr<StreamSocket> socket = _socket.lock();
+    std::shared_ptr<StreamSocket> socket = getSocket().lock();
     if (socket)
     {
         uint64_t sent, recv;
@@ -954,4 +1226,142 @@ void ClientSession::dumpState(std::ostream& os)
 
 }
 
+void ClientSession::handleTileInvalidation(const std::string& message,
+    const std::shared_ptr<DocumentBroker>& docBroker)
+{
+    docBroker->invalidateTiles(message);
+
+    // Skip requesting new tiles if we don't have client visible area data yet.
+    if(!_clientVisibleArea.hasSurface() ||
+       _tileWidthPixel == 0 || _tileHeightPixel == 0 ||
+       _tileWidthTwips == 0 || _tileHeightTwips == 0 ||
+       (_clientSelectedPart == -1 && !_isTextDocument))
+    {
+        return;
+    }
+
+    // Visible area can have negativ value as position, but we have tiles only in the positive range
+    Util::Rectangle normalizedVisArea = getNormalizedVisibleArea();
+
+    std::pair<int, Util::Rectangle> result = TileCache::parseInvalidateMsg(message);
+    int part = result.first;
+    Util::Rectangle& invalidateRect = result.second;
+
+    // We can ignore the invalidation if it's outside of the visible area
+    if(!normalizedVisArea.intersects(invalidateRect))
+        return;
+
+    if( part == -1 ) // If no part is specifed we use the part used by the client
+        part = _clientSelectedPart;
+
+
+    std::vector<TileDesc> invalidTiles;
+    if(part == _clientSelectedPart || _isTextDocument)
+    {
+        // Iterate through visible tiles
+        for(int i = std::ceil(normalizedVisArea.getTop() / _tileHeightTwips);
+                    i <= std::ceil(normalizedVisArea.getBottom() / _tileHeightTwips); ++i)
+        {
+            for(int j = std::ceil(normalizedVisArea.getLeft() / _tileWidthTwips);
+                j <= std::ceil(normalizedVisArea.getRight() / _tileWidthTwips); ++j)
+            {
+                // Find tiles affected by invalidation
+                Util::Rectangle tileRect (j * _tileWidthTwips, i * _tileHeightTwips, _tileWidthTwips, _tileHeightTwips);
+                if(invalidateRect.intersects(tileRect))
+                {
+                    invalidTiles.emplace_back(part, _tileWidthPixel, _tileHeightPixel, j * _tileWidthTwips, i * _tileHeightTwips, _tileWidthTwips, _tileHeightTwips, -1, 0, -1, false);
+
+                    TileWireId oldWireId = 0;
+                    auto iter = _oldWireIds.find(generateTileID(invalidTiles.back()));
+                    if(iter != _oldWireIds.end())
+                        oldWireId = iter->second;
+
+                    invalidTiles.back().setOldWireId(oldWireId);
+                    invalidTiles.back().setWireId(0);
+                }
+            }
+        }
+    }
+
+    if(!invalidTiles.empty())
+    {
+        TileCombined tileCombined = TileCombined::create(invalidTiles);
+        docBroker->handleTileCombinedRequest(tileCombined, shared_from_this());
+    }
+}
+
+void ClientSession::resetWireIdMap()
+{
+    _oldWireIds.clear();
+}
+
+void ClientSession::traceTileBySend(const TileDesc& tile, bool deduplicated)
+{
+    const std::string tileID = generateTileID(tile);
+
+    // Store wireId first
+    auto iter = _oldWireIds.find(tileID);
+    if(iter != _oldWireIds.end())
+    {
+        iter->second = tile.getWireId();
+    }
+    else
+    {
+        // Track only tile inside the visible area
+        if(_clientVisibleArea.hasSurface() &&
+           tile.getTilePosX() >= _clientVisibleArea.getLeft() && tile.getTilePosX() <= _clientVisibleArea.getRight() &&
+           tile.getTilePosY() >= _clientVisibleArea.getTop() && tile.getTilePosY() <= _clientVisibleArea.getBottom())
+        {
+            _oldWireIds.insert(std::pair<std::string, TileWireId>(tileID, tile.getWireId()));
+        }
+    }
+
+    // Record that the tile is sent
+    if (!deduplicated)
+        addTileOnFly(tile);
+}
+
+void ClientSession::traceSubscribeToTile(const std::string& cacheName)
+{
+    _tilesBeingRendered.insert(cacheName);
+}
+
+void ClientSession::traceUnSubscribeToTile(const std::string& cacheName)
+{
+    _tilesBeingRendered.erase(cacheName);
+}
+
+void ClientSession::removeOutdatedTileSubscriptions()
+{
+    const std::shared_ptr<DocumentBroker> docBroker = getDocumentBroker();
+    if(!docBroker)
+        return;
+
+    auto iterator = _tilesBeingRendered.begin();
+    while(iterator != _tilesBeingRendered.end())
+    {
+        double elapsedTime = docBroker->tileCache().getTileBeingRenderedElapsedTimeMs(*iterator);
+        if(elapsedTime < 0.0 && elapsedTime > 200.0)
+        {
+            LOG_INF("Tracked TileBeingRendered was dropped because of time out.");
+            _tilesBeingRendered.erase(iterator);
+        }
+        else
+            ++iterator;
+    }
+}
+
+void ClientSession::clearTileSubscription()
+{
+    _tilesBeingRendered.clear();
+}
+
+std::string ClientSession::generateTileID(const TileDesc& tile) const
+{
+    std::ostringstream tileID;
+    tileID << tile.getPart() << ":" << tile.getTilePosX() << ":" << tile.getTilePosY() << ":"
+           << tile.getTileWidth() << ":" << tile.getTileHeight();
+    return tileID.str();
+}
+
 /* vim:set shiftwidth=4 softtabstop=4 expandtab: */
