@@ -37,11 +37,12 @@
 #include <common/Message.hpp>
 #include <common/Protocol.hpp>
 #include <common/Unit.hpp>
-#include <common/FileUtil.hpp>
 
 #include <sys/types.h>
 #include <sys/wait.h>
 
+#define TILES_ON_FLY_MIN_UPPER_LIMIT 10.0f
+
 using namespace LOOLProtocol;
 
 using Poco::JSON::Object;
@@ -144,11 +145,13 @@ std::atomic<unsigned> DocumentBroker::DocBrokerId(1);
 
 DocumentBroker::DocumentBroker(const std::string& uri,
                                const Poco::URI& uriPublic,
-                               const std::string& docKey) :
+                               const std::string& docKey,
+                               const std::string& childRoot) :
     _uriOrig(uri),
     _uriPublic(uriPublic),
     _docKey(docKey),
     _docId(Util::encodeId(DocBrokerId++, 3)),
+    _childRoot(childRoot),
     _cacheRoot(getCachePath(uriPublic.toString())),
     _documentChangedInStorage(false),
     _lastSaveTime(std::chrono::steady_clock::now()),
@@ -168,10 +171,10 @@ DocumentBroker::DocumentBroker(const std::string& uri,
     _debugRenderedTileCount(0)
 {
     assert(!_docKey.empty());
-    assert(!LOOLWSD::ChildRoot.empty());
+    assert(!_childRoot.empty());
 
-    LOG_INF("DocumentBroker [" << _uriPublic.toString() <<
-            "] created with docKey [" << _docKey << "]");
+    LOG_INF("DocumentBroker [" << LOOLWSD::anonymizeUrl(_uriPublic.toString()) <<
+            "] created with docKey [" << _docKey << "] and root [" << _childRoot << "]");
 }
 
 void DocumentBroker::startThread()
@@ -192,6 +195,7 @@ void DocumentBroker::pollThread()
     _threadStart = std::chrono::steady_clock::now();
 
     // Request a kit process for this doc.
+#ifndef MOBILEAPP
     do
     {
         static const int timeoutMs = COMMAND_TIMEOUT_MS * 5;
@@ -200,11 +204,13 @@ void DocumentBroker::pollThread()
             std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::steady_clock::now() -
                                                                   _threadStart).count() > timeoutMs)
             break;
-
         // Nominal time between retries, lest we busy-loop. getNewChild could also wait, so don't double that here.
         std::this_thread::sleep_for(std::chrono::milliseconds(CHILD_REBALANCE_INTERVAL_MS / 10));
     }
     while (!_stop && _poll->continuePolling() && !TerminationFlag && !ShutdownRequestFlag);
+#else
+    _childProcess = getNewChild_Blocks(getPublicUri().getPath());
+#endif
 
     if (!_childProcess)
     {
@@ -239,12 +245,17 @@ void DocumentBroker::pollThread()
     static const size_t IdleDocTimeoutSecs = LOOLWSD::getConfigValue<int>(
                                                       "per_document.idle_timeout_secs", 3600);
 
+#ifndef MOBILEAPP
     // Used to accumulate B/W deltas.
     uint64_t adminSent = 0;
     uint64_t adminRecv = 0;
     auto lastBWUpdateTime = std::chrono::steady_clock::now();
+#endif
     auto last30SecCheckTime = std::chrono::steady_clock::now();
 
+    int limit_load_secs = LOOLWSD::getConfigValue<int>("per_document.limit_load_secs", 100);
+    auto loadDeadline = std::chrono::steady_clock::now() + std::chrono::seconds(limit_load_secs);
+
     // Main polling loop goodness.
     while (!_stop && _poll->continuePolling() && !TerminationFlag)
     {
@@ -252,6 +263,16 @@ void DocumentBroker::pollThread()
 
         const auto now = std::chrono::steady_clock::now();
 
+#ifndef MOBILEAPP
+        if (!_isLoaded && (limit_load_secs > 0) && (now > loadDeadline))
+        {
+            // Brutal but effective.
+            if (_childProcess)
+                _childProcess->terminate();
+            stop("Load timed out");
+            continue;
+        }
+
         if (std::chrono::duration_cast<std::chrono::milliseconds>
                     (now - lastBWUpdateTime).count() >= 5 * 1000)
         {
@@ -267,6 +288,7 @@ void DocumentBroker::pollThread()
             adminSent = sent;
             adminRecv = recv;
         }
+#endif
 
         if (isSaving() &&
             std::chrono::duration_cast<std::chrono::milliseconds>
@@ -349,8 +371,10 @@ void DocumentBroker::pollThread()
     _poll->stop();
     _poll->removeSockets();
 
+#ifndef MOBILEAPP
     // Async cleanup.
     LOOLWSD::doHousekeeping();
+#endif
 
     // Remove all tiles related to this document from the cache if configured so.
     if (_tileCache && !LOOLWSD::TileCachePersistent)
@@ -372,7 +396,9 @@ DocumentBroker::~DocumentBroker()
 {
     assertCorrectThread();
 
+#ifndef MOBILEAPP
     Admin::instance().rmDoc(_docKey);
+#endif
 
     LOG_INF("~DocumentBroker [" << _docKey <<
             "] destroyed with " << _sessions.size() << " sessions left.");
@@ -441,7 +467,7 @@ bool DocumentBroker::load(const std::shared_ptr<ClientSession>& session, const s
         // Pass the public URI to storage as it needs to load using the token
         // and other storage-specific data provided in the URI.
         const Poco::URI& uriPublic = session->getPublicUri();
-        LOG_DBG("Loading, and creating new storage instance for URI [" << uriPublic.toString() << "].");
+        LOG_DBG("Loading, and creating new storage instance for URI [" << LOOLWSD::anonymizeUrl(uriPublic.toString()) << "].");
 
         _storage = StorageBase::create(uriPublic, jailRoot, jailPath.toString());
         if (_storage == nullptr)
@@ -456,15 +482,18 @@ bool DocumentBroker::load(const std::shared_ptr<ClientSession>& session, const s
     assert(_storage != nullptr);
 
     // Call the storage specific fileinfo functions
-    std::string userid, username;
+    std::string userId, username;
     std::string userExtraInfo;
     std::string watermarkText;
+
+#ifndef MOBILEAPP
     std::chrono::duration<double> getInfoCallDuration(0);
+
     WopiStorage* wopiStorage = dynamic_cast<WopiStorage*>(_storage.get());
     if (wopiStorage != nullptr)
     {
         std::unique_ptr<WopiStorage::WOPIFileInfo> wopifileinfo = wopiStorage->getWOPIFileInfo(session->getAuthorization());
-        userid = wopifileinfo->_userid;
+        userId = wopifileinfo->_userId;
         username = wopifileinfo->_username;
         userExtraInfo = wopifileinfo->_userExtraInfo;
         watermarkText = wopifileinfo->_watermarkText;
@@ -501,22 +530,28 @@ bool DocumentBroker::load(const std::shared_ptr<ClientSession>& session, const s
         wopiInfo->set("HidePrintOption", wopifileinfo->_hidePrintOption);
         wopiInfo->set("HideSaveOption", wopifileinfo->_hideSaveOption);
         wopiInfo->set("HideExportOption", wopifileinfo->_hideExportOption);
-        wopiInfo->set("HideChangeTrackingControls", wopifileinfo->_hideChangeTrackingControls);
         wopiInfo->set("DisablePrint", wopifileinfo->_disablePrint);
         wopiInfo->set("DisableExport", wopifileinfo->_disableExport);
         wopiInfo->set("DisableCopy", wopifileinfo->_disableCopy);
         wopiInfo->set("DisableInactiveMessages", wopifileinfo->_disableInactiveMessages);
         wopiInfo->set("UserCanNotWriteRelative", wopifileinfo->_userCanNotWriteRelative);
+        wopiInfo->set("EnableInsertRemoteImage", wopifileinfo->_enableInsertRemoteImage);
+        wopiInfo->set("EnableShare", wopifileinfo->_enableShare);
+        if (wopifileinfo->_hideChangeTrackingControls != WopiStorage::WOPIFileInfo::TriState::Unset)
+            wopiInfo->set("HideChangeTrackingControls", wopifileinfo->_hideChangeTrackingControls == WopiStorage::WOPIFileInfo::TriState::True);
 
         std::ostringstream ossWopiInfo;
         wopiInfo->stringify(ossWopiInfo);
+        const std::string wopiInfoString = ossWopiInfo.str();
+        LOG_TRC("Sending wopi info to client: " << wopiInfoString);
+
         // Contains PostMessageOrigin property which is necessary to post messages to parent
         // frame. Important to send this message immediately and not enqueue it so that in case
         // document load fails, loleaflet is able to tell its parent frame via PostMessage API.
-        session->sendMessage("wopi: " + ossWopiInfo.str());
+        session->sendMessage("wopi: " + wopiInfoString);
 
         // Mark the session as 'Document owner' if WOPI hosts supports it
-        if (userid == _storage->getFileInfo()._ownerId)
+        if (userId == _storage->getFileInfo()._ownerId)
         {
             LOG_DBG("Session [" << sessionId << "] is the document owner");
             session->setDocumentOwner(true);
@@ -528,12 +563,13 @@ bool DocumentBroker::load(const std::shared_ptr<ClientSession>& session, const s
         session->setWopiFileInfo(wopifileinfo);
     }
     else
+#endif
     {
         LocalStorage* localStorage = dynamic_cast<LocalStorage*>(_storage.get());
         if (localStorage != nullptr)
         {
             std::unique_ptr<LocalStorage::LocalFileInfo> localfileinfo = localStorage->getLocalFileInfo();
-            userid = localfileinfo->_userid;
+            userId = localfileinfo->_userId;
             username = localfileinfo->_username;
 
             if (LOOLWSD::IsViewFileExtension(localStorage->getFileExtension()))
@@ -544,14 +580,16 @@ bool DocumentBroker::load(const std::shared_ptr<ClientSession>& session, const s
         }
     }
 
+
 #if ENABLE_SUPPORT_KEY
     if (!LOOLWSD::OverrideWatermark.empty())
         watermarkText = LOOLWSD::OverrideWatermark;
 #endif
 
-    LOG_DBG("Setting username [" << username << "] and userId [" <<
-            userid << "] for session [" << sessionId << "]");
-    session->setUserId(userid);
+    LOG_DBG("Setting username [" << LOOLWSD::anonymizeUsername(username) << "] and userId [" <<
+            LOOLWSD::anonymizeUsername(userId) << "] for session [" << sessionId << "]");
+
+    session->setUserId(userId);
     session->setUserName(username);
     session->setUserExtraInfo(userExtraInfo);
     session->setWatermarkText(watermarkText);
@@ -578,7 +616,7 @@ bool DocumentBroker::load(const std::shared_ptr<ClientSession>& session, const s
             fileInfo._modifiedTime != Zero &&
             _documentLastModifiedTime != fileInfo._modifiedTime)
         {
-            LOG_TRC("Document " << _docKey << "] has been modified behind our back. " <<
+            LOG_DBG("Document " << _docKey << "] has been modified behind our back. " <<
                     "Informing all clients. Expected: " << _documentLastModifiedTime <<
                     ", Actual: " << fileInfo._modifiedTime);
 
@@ -597,6 +635,7 @@ bool DocumentBroker::load(const std::shared_ptr<ClientSession>& session, const s
     {
         std::string localPath = _storage->loadStorageFileToLocal(session->getAuthorization());
 
+#ifndef MOBILEAPP
         // Check if we have a prefilter "plugin" for this document format
         for (const auto& plugin : LOOLWSD::PluginConfigurations)
         {
@@ -648,28 +687,31 @@ bool DocumentBroker::load(const std::shared_ptr<ClientSession>& session, const s
                 // This plugin is not a proper prefilter one
             }
         }
+#endif
 
         std::ifstream istr(localPath, std::ios::binary);
         Poco::SHA1Engine sha1;
         Poco::DigestOutputStream dos(sha1);
         Poco::StreamCopier::copyStream(istr, dos);
         dos.close();
-        LOG_INF("SHA1 for DocKey [" << _docKey << "] of [" << localPath << "]: " <<
+        LOG_INF("SHA1 for DocKey [" << _docKey << "] of [" << LOOLWSD::anonymizeUrl(localPath) << "]: " <<
                 Poco::DigestEngine::digestToHex(sha1.digest()));
 
         // LibreOffice can't open files with '#' in the name
         std::string localPathEncoded;
-        Poco::URI::encode(localPath,"#",localPathEncoded);
-        _uriJailed = Poco::URI(Poco::URI("file://"), localPathEncoded);
+        Poco::URI::encode(localPath, "#", localPathEncoded);
+        _uriJailed = Poco::URI(Poco::URI("file://"), localPathEncoded).toString();
+        _uriJailedAnonym = Poco::URI(Poco::URI("file://"), LOOLWSD::anonymizeUrl(localPath)).toString();
 
         _filename = fileInfo._filename;
 
         // Use the local temp file's timestamp.
         _lastFileModifiedTime = Poco::File(_storage->getRootFilePath()).getLastModified();
-        _tileCache.reset(new TileCache(_storage->getUri(), _lastFileModifiedTime, _cacheRoot));
+        _tileCache.reset(new TileCache(_storage->getUri(), _lastFileModifiedTime, _cacheRoot, LOOLWSD::TileCachePersistent));
         _tileCache->setThreadOwner(std::this_thread::get_id());
     }
 
+#ifndef MOBILEAPP
     LOOLWSD::dumpNewSessionTrace(getJailId(), sessionId, _uriOrig, _storage->getRootFilePath());
 
     // Since document has been loaded, send the stats if its WOPI
@@ -683,7 +725,7 @@ bool DocumentBroker::load(const std::shared_ptr<ClientSession>& session, const s
         LOG_TRC("Sending to Client [" << msg << "].");
         session->sendTextFrame(msg);
     }
-
+#endif
     return true;
 }
 
@@ -761,20 +803,29 @@ bool DocumentBroker::saveToStorageInternal(const std::string& sessionId,
     }
 
     const Authorization auth = it->second->getAuthorization();
-    const std::string uri = isSaveAs? saveAsPath: it->second->getPublicUri().toString();
+    const std::string uri = isSaveAs ? saveAsPath : it->second->getPublicUri().toString();
+
+    // Map the FileId from the docKey to the new filename to anonymize the new filename as the FileId.
+    const std::string newFilename = Util::getFilenameFromURL(uri);
+    const std::string fileId = Util::getFilenameFromURL(_docKey);
+    if (LOOLWSD::AnonymizeFilenames)
+        LOG_DBG("New filename [" << LOOLWSD::anonymizeUrl(newFilename) << "] will be known by its fileId [" << fileId << "]");
+
+    Util::mapAnonymized(newFilename, fileId);
+    const std::string uriAnonym = LOOLWSD::anonymizeUrl(uri);
 
     // If the file timestamp hasn't changed, skip saving.
     const Poco::Timestamp newFileModifiedTime = Poco::File(_storage->getRootFilePath()).getLastModified();
     if (!isSaveAs && newFileModifiedTime == _lastFileModifiedTime)
     {
         // Nothing to do.
-        LOG_DBG("Skipping unnecessary saving to URI [" << uri << "] with docKey [" << _docKey <<
+        LOG_DBG("Skipping unnecessary saving to URI [" << uriAnonym << "] with docKey [" << _docKey <<
                 "]. File last modified " << _lastFileModifiedTime.elapsed() / 1000000 << " seconds ago.");
         _poll->wakeup();
         return true;
     }
 
-    LOG_DBG("Persisting [" << _docKey << "] after saving to URI [" << uri << "].");
+    LOG_DBG("Persisting [" << _docKey << "] after saving to URI [" << uriAnonym << "].");
 
     assert(_storage && _tileCache);
     StorageBase::SaveResult storageSaveResult = _storage->saveLocalFileToStorage(auth, saveAsPath, saveAsFilename);
@@ -794,7 +845,7 @@ bool DocumentBroker::saveToStorageInternal(const std::string& sessionId,
             // After a successful save, we are sure that document in the storage is same as ours
             _documentChangedInStorage = false;
 
-            LOG_DBG("Saved docKey [" << _docKey << "] to URI [" << uri << "] and updated timestamps. " <<
+            LOG_DBG("Saved docKey [" << _docKey << "] to URI [" << uriAnonym << "] and updated timestamps. " <<
                     " Document modified timestamp: " << _documentLastModifiedTime);
 
             // Resume polling.
@@ -803,22 +854,29 @@ bool DocumentBroker::saveToStorageInternal(const std::string& sessionId,
         else
         {
             // normalize the url (mainly to " " -> "%20")
-            std::string url = Poco::URI(storageSaveResult.getSaveAsUrl()).toString();
+            const std::string url = Poco::URI(storageSaveResult.getSaveAsUrl()).toString();
+
+            const std::string filename = storageSaveResult.getSaveAsName();
 
             // encode the name
             std::string encodedName;
-            Poco::URI::encode(storageSaveResult.getSaveAsName(), "", encodedName);
+            Poco::URI::encode(filename, "", encodedName);
+            const std::string filenameAnonym = LOOLWSD::anonymizeUrl(filename);
 
-            it->second->sendTextFrame("saveas: url=" + url + " filename=" + encodedName);
+            std::ostringstream oss;
+            oss << "saveas: url=" << url << " filename=" << encodedName
+                << " xfilename=" << filenameAnonym;
+            it->second->sendTextFrame(oss.str());
 
-            LOG_DBG("Saved As docKey [" << _docKey << "] to URI [" << url <<
-                    " with name '" << encodedName << "'] successfully.");
+            LOG_DBG("Saved As docKey [" << _docKey << "] to URI [" << LOOLWSD::anonymizeUrl(url) <<
+                    "] with name [" << filenameAnonym << "] successfully.");
         }
+
         return true;
     }
     else if (storageSaveResult.getResult() == StorageBase::SaveResult::DISKFULL)
     {
-        LOG_WRN("Disk full while saving docKey [" << _docKey << "] to URI [" << uri <<
+        LOG_WRN("Disk full while saving docKey [" << _docKey << "] to URI [" << uriAnonym <<
                 "]. Making all sessions on doc read-only and notifying clients.");
 
         // Make everyone readonly and tell everyone that storage is low on diskspace.
@@ -830,13 +888,14 @@ bool DocumentBroker::saveToStorageInternal(const std::string& sessionId,
     }
     else if (storageSaveResult.getResult() == StorageBase::SaveResult::UNAUTHORIZED)
     {
-        LOG_ERR("Cannot save docKey [" << _docKey << "] to storage URI [" << uri << "]. Invalid or expired access token. Notifying client.");
+        LOG_ERR("Cannot save docKey [" << _docKey << "] to storage URI [" << uriAnonym <<
+                "]. Invalid or expired access token. Notifying client.");
         it->second->sendTextFrame("error: cmd=storage kind=saveunauthorized");
     }
     else if (storageSaveResult.getResult() == StorageBase::SaveResult::FAILED)
     {
         //TODO: Should we notify all clients?
-        LOG_ERR("Failed to save docKey [" << _docKey << "] to URI [" << uri << "]. Notifying client.");
+        LOG_ERR("Failed to save docKey [" << _docKey << "] to URI [" << uriAnonym << "]. Notifying client.");
         it->second->sendTextFrame("error: cmd=storage kind=savefailed");
     }
     else if (storageSaveResult.getResult() == StorageBase::SaveResult::DOC_CHANGED)
@@ -903,7 +962,9 @@ bool DocumentBroker::autoSave(const bool force)
     {
         LOG_TRC("Sending forced save command for [" << _docKey << "].");
         // Don't terminate editing as this can be invoked by the admin OOM, but otherwise force saving anyway.
-        sent = sendUnoSave(savingSessionId, /*dontTerminateEdit=*/ true, /*dontSaveIfUnmodified=*/ true, /*isAutosave=*/ false);
+        sent = sendUnoSave(savingSessionId, /*dontTerminateEdit=*/true,
+                           /*dontSaveIfUnmodified=*/true, /*isAutosave=*/false,
+                           /*isExitSave=*/true);
     }
     else if (_isModified)
     {
@@ -920,14 +981,17 @@ bool DocumentBroker::autoSave(const bool force)
             timeSinceLastSaveMs >= autoSaveDurationMs)
         {
             LOG_TRC("Sending timed save command for [" << _docKey << "].");
-            sent = sendUnoSave(savingSessionId, /*dontTerminateEdit=*/ true, /*dontSaveIfUnmodified=*/ true, /*isAutosave=*/ true);
+            sent = sendUnoSave(savingSessionId, /*dontTerminateEdit=*/true,
+                               /*dontSaveIfUnmodified=*/true, /*isAutosave=*/true,
+                               /*isExitSave=*/false);
         }
     }
 
     return sent;
 }
 
-bool DocumentBroker::sendUnoSave(const std::string& sessionId, bool dontTerminateEdit, bool dontSaveIfUnmodified, bool isAutosave)
+bool DocumentBroker::sendUnoSave(const std::string& sessionId, bool dontTerminateEdit,
+                                 bool dontSaveIfUnmodified, bool isAutosave, bool isExitSave)
 {
     assertCorrectThread();
 
@@ -970,6 +1034,7 @@ bool DocumentBroker::sendUnoSave(const std::string& sessionId, bool dontTerminat
 
         assert(_storage);
         _storage->setIsAutosave(isAutosave || UnitWSD::get().isAutosave());
+        _storage->setIsExitSave(isExitSave);
 
         const std::string saveArgs = oss.str();
         LOG_TRC(".uno:Save arguments: " << saveArgs);
@@ -986,7 +1051,7 @@ bool DocumentBroker::sendUnoSave(const std::string& sessionId, bool dontTerminat
 std::string DocumentBroker::getJailRoot() const
 {
     assert(!_jailId.empty());
-    return Poco::Path(LOOLWSD::ChildRoot, _jailId).toString();
+    return Poco::Path(_childRoot, _jailId).toString();
 }
 
 size_t DocumentBroker::addSession(const std::shared_ptr<ClientSession>& session)
@@ -997,7 +1062,7 @@ size_t DocumentBroker::addSession(const std::shared_ptr<ClientSession>& session)
     }
     catch (const std::exception& exc)
     {
-        LOG_ERR("Failed to add session to [" << _docKey << "] with URI [" << session->getPublicUri().toString() << "]: " << exc.what());
+        LOG_ERR("Failed to add session to [" << _docKey << "] with URI [" << LOOLWSD::anonymizeUrl(session->getPublicUri().toString()) << "]: " << exc.what());
         if (_sessions.empty())
         {
             LOG_INF("Doc [" << _docKey << "] has no more sessions. Marking to destroy.");
@@ -1040,8 +1105,10 @@ size_t DocumentBroker::addSessionInternal(const std::shared_ptr<ClientSession>&
     const std::string aMessage = "session " + id + ' ' + _docKey + ' ' + _docId;
     _childProcess->sendTextFrame(aMessage);
 
+#ifndef MOBILEAPP
     // Tell the admin console about this new doc
     Admin::instance().addDoc(_docKey, getPid(), getFilename(), id, session->getUserName(), session->getUserId());
+#endif
 
     // Add and attach the session.
     _sessions.emplace(session->getId(), session);
@@ -1094,12 +1161,15 @@ size_t DocumentBroker::removeSessionInternal(const std::string& id)
     assertCorrectThread();
     try
     {
+#ifndef MOBILEAPP
         Admin::instance().rmDoc(_docKey, id);
-
+#endif
         auto it = _sessions.find(id);
         if (it != _sessions.end())
         {
+#ifndef MOBILEAPP
             LOOLWSD::dumpEndSessionTrace(getJailId(), id, _uriOrig);
+#endif
 
             const bool readonly = (it->second ? it->second->isReadOnly() : false);
 
@@ -1117,7 +1187,7 @@ size_t DocumentBroker::removeSessionInternal(const std::string& id)
                 for (const auto& pair : _sessions)
                     logger << pair.second->getId() << ' ';
 
-                LOG_END(logger);
+                LOG_END(logger, true);
             }
 
             // Let the child know the client has disconnected.
@@ -1173,7 +1243,9 @@ bool DocumentBroker::handleInput(const std::vector<char>& payload)
     const auto& msg = message->abbr();
     LOG_TRC("DocumentBroker handling child message: [" << msg << "].");
 
+#ifndef MOBILEAPP
     LOOLWSD::dumpOutgoingTrace(getJailId(), "0", msg);
+#endif
 
     if (LOOLProtocol::getFirstToken(message->forwardToken(), '-') == "client")
     {
@@ -1200,6 +1272,7 @@ bool DocumentBroker::handleInput(const std::vector<char>& payload)
             LOG_CHECK_RET(kind != "", false);
             Util::alertAllUsers(cmd, kind);
         }
+#ifndef MOBILEAPP
         else if (command == "procmemstats:")
         {
             int dirty;
@@ -1208,6 +1281,7 @@ bool DocumentBroker::handleInput(const std::vector<char>& payload)
                 Admin::instance().updateMemoryDirty(_docKey, dirty);
             }
         }
+#endif
         else
         {
             LOG_ERR("Unexpected message: [" << msg << "].");
@@ -1288,61 +1362,202 @@ void DocumentBroker::handleTileCombinedRequest(TileCombined& tileCombined,
 
     LOG_TRC("TileCombined request for " << tileCombined.serialize());
 
-    // Satisfy as many tiles from the cache.
-    std::vector<TileDesc> tiles;
+    // Check which newly requested tiles needs rendering.
+    std::vector<TileDesc> tilesNeedsRendering;
     for (auto& tile : tileCombined.getTiles())
     {
+        tile.setVersion(++_tileVersion);
         std::unique_ptr<std::fstream> cachedTile = _tileCache->lookupTile(tile);
-        if (cachedTile)
-        {
-            //TODO: Combine the response to reduce latency.
-#if ENABLE_DEBUG
-            const std::string response = tile.serialize("tile:") + " renderid=cached\n";
-#else
-            const std::string response = tile.serialize("tile:") + "\n";
-#endif
-
-            std::vector<char> output;
-            output.reserve(static_cast<size_t>(4) * tile.getWidth() * tile.getHeight());
-            output.resize(response.size());
-            std::memcpy(output.data(), response.data(), response.size());
-
-            assert(cachedTile->is_open());
-            cachedTile->seekg(0, std::ios_base::end);
-            const size_t pos = output.size();
-            std::streamsize size = cachedTile->tellg();
-            output.resize(pos + size);
-            cachedTile->seekg(0, std::ios_base::beg);
-            cachedTile->read(output.data() + pos, size);
+        if(cachedTile)
             cachedTile->close();
-
-            session->sendBinaryFrame(output.data(), output.size());
-        }
         else
         {
             // Not cached, needs rendering.
-            tile.setVersion(++_tileVersion);
-            tileCache().subscribeToTileRendering(tile, session);
-            tiles.push_back(tile);
+            tilesNeedsRendering.push_back(tile);
             _debugRenderedTileCount++;
+            tileCache().registerTileBeingRendered(tile);
         }
     }
 
-    if (!tiles.empty())
+    // Send rendering request, prerender before we actually send the tiles
+    if (!tilesNeedsRendering.empty())
     {
-        TileCombined newTileCombined = TileCombined::create(tiles);
+        TileCombined newTileCombined = TileCombined::create(tilesNeedsRendering);
 
         // Forward to child to render.
         const std::string req = newTileCombined.serialize("tilecombine");
         LOG_DBG("Sending residual tilecombine: " << req);
         _childProcess->sendTextFrame(req);
     }
+
+    // Accumulate tiles
+    std::deque<TileDesc>& requestedTiles = session->getRequestedTiles();
+    if (requestedTiles.empty())
+    {
+        requestedTiles = std::deque<TileDesc>(tileCombined.getTiles().begin(), tileCombined.getTiles().end());
+    }
+    // Drop duplicated tiles, but use newer version number
+    else
+    {
+        for (const auto& newTile : tileCombined.getTiles())
+        {
+            const TileDesc& firstOldTile = *(requestedTiles.begin());
+            if(!session->isTextDocument() && newTile.getPart() != firstOldTile.getPart())
+            {
+                LOG_WRN("Different part numbers in tile requests");
+            }
+            else if (newTile.getTileWidth() != firstOldTile.getTileWidth() ||
+                     newTile.getTileHeight() != firstOldTile.getTileHeight() )
+            {
+                LOG_WRN("Different tile sizes in tile requests");
+            }
+
+            bool tileFound = false;
+            for (auto& oldTile : requestedTiles)
+            {
+                if(oldTile.getTilePosX() == newTile.getTilePosX() &&
+                   oldTile.getTilePosY() == newTile.getTilePosY() )
+                {
+                    oldTile.setVersion(newTile.getVersion());
+                    oldTile.setOldWireId(newTile.getOldWireId());
+                    oldTile.setWireId(newTile.getWireId());
+                    tileFound = true;
+                    break;
+                }
+            }
+            if(!tileFound)
+                requestedTiles.push_back(newTile);
+        }
+    }
+
+    lock.unlock();
+    lock.release();
+    sendRequestedTiles(session);
+}
+
+void DocumentBroker::sendRequestedTiles(const std::shared_ptr<ClientSession>& session)
+{
+    std::unique_lock<std::mutex> lock(_mutex);
+
+    // How many tiles we have on the visible area, set the upper limit accordingly
+    Util::Rectangle normalizedVisArea = session->getNormalizedVisibleArea();
+
+    float tilesOnFlyUpperLimit = 0;
+    if (normalizedVisArea.hasSurface() && session->getTileWidthInTwips() != 0 && session->getTileHeightInTwips() != 0)
+    {
+
+        const int tilesFitOnWidth = std::ceil(normalizedVisArea.getRight() / session->getTileWidthInTwips()) -
+                                    std::ceil(normalizedVisArea.getLeft() / session->getTileWidthInTwips()) + 1;
+        const int tilesFitOnHeight = std::ceil(normalizedVisArea.getBottom() / session->getTileHeightInTwips()) -
+                                     std::ceil(normalizedVisArea.getTop() / session->getTileHeightInTwips()) + 1;
+        const int tilesInVisArea = tilesFitOnWidth * tilesFitOnHeight;
+
+        tilesOnFlyUpperLimit = std::max(TILES_ON_FLY_MIN_UPPER_LIMIT, tilesInVisArea * 1.1f);
+    }
+    else
+    {
+        tilesOnFlyUpperLimit = 200; // Have a big number here to get all tiles requested by file openning
+    }
+
+
+    // Update client's tilesBeingRendered list
+    session->removeOutdatedTileSubscriptions();
+    // Drop tiles which we are waiting for too long
+    session->removeOutdatedTilesOnFly();
+
+    // All tiles were processed on client side what we sent last time, so we can send a new banch of tiles
+    // which was invalidated / requested in the meantime
+    std::deque<TileDesc>& requestedTiles = session->getRequestedTiles();
+    if (!requestedTiles.empty())
+    {
+        size_t delayedTiles = 0;
+        std::vector<TileDesc> tilesNeedsRendering;
+        while(session->getTilesOnFlyCount() + session->getTilesBeingRenderedCount() < tilesOnFlyUpperLimit &&
+              !requestedTiles.empty() &&
+              // If we delayed all tiles we don't send any tile (we will when next tileprocessed message arrives)
+              delayedTiles < requestedTiles.size())
+        {
+            TileDesc& tile = *(requestedTiles.begin());
+
+            // We already sent out two versions of the same tile, let's not send the third one
+            // until we get a tileprocessed message for this specific tile.
+            if (session->countIdenticalTilesOnFly(tile) >= 2)
+            {
+                requestedTiles.push_back(requestedTiles.front());
+                requestedTiles.pop_front();
+                delayedTiles += 1;
+                LOG_INF("Requested tile was delayed!");
+                continue;
+            }
+
+            // Satisfy as many tiles from the cache.
+            std::unique_ptr<std::fstream> cachedTile = _tileCache->lookupTile(tile);
+            if (cachedTile)
+            {
+                //TODO: Combine the response to reduce latency.
+#if ENABLE_DEBUG
+                const std::string response = tile.serialize("tile:") + " renderid=cached\n";
+#else
+                const std::string response = tile.serialize("tile:") + "\n";
+#endif
+
+                std::vector<char> output;
+                output.reserve(static_cast<size_t>(4) * tile.getWidth() * tile.getHeight());
+                output.resize(response.size());
+                std::memcpy(output.data(), response.data(), response.size());
+
+                assert(cachedTile->is_open());
+                cachedTile->seekg(0, std::ios_base::end);
+                const auto pos = output.size();
+                std::streamsize size = cachedTile->tellg();
+                output.resize(pos + size);
+                cachedTile->seekg(0, std::ios_base::beg);
+                cachedTile->read(output.data() + pos, size);
+                cachedTile->close();
+
+                session->sendBinaryFrame(output.data(), output.size());
+            }
+            else
+            {
+                // Not cached, needs rendering.
+                if (!tileCache().hasTileBeingRendered(tile) || // There is no in progress rendering of the given tile
+                    tileCache().getTileBeingRenderedVersion(tile) < tile.getVersion()) // We need a newer version
+                {
+                    tile.setVersion(++_tileVersion);
+                    tilesNeedsRendering.push_back(tile);
+                    _debugRenderedTileCount++;
+                }
+                tileCache().subscribeToTileRendering(tile, session);
+            }
+            requestedTiles.pop_front();
+        }
+
+        // Send rendering request for those tiles which were not prerendered
+        if (!tilesNeedsRendering.empty())
+        {
+            TileCombined newTileCombined = TileCombined::create(tilesNeedsRendering);
+
+            // Forward to child to render.
+            const std::string req = newTileCombined.serialize("tilecombine");
+            LOG_DBG("Some of the tiles were not prerendered. Sending residual tilecombine: " << req);
+            _childProcess->sendTextFrame(req);
+        }
+    }
 }
 
 void DocumentBroker::cancelTileRequests(const std::shared_ptr<ClientSession>& session)
 {
     std::unique_lock<std::mutex> lock(_mutex);
 
+    // Clear tile requests
+    session->clearTilesOnFly();
+
+    session->getRequestedTiles().clear();
+
+    session->clearTileSubscription();
+
+    session->resetWireIdMap();
+
     const std::string canceltiles = tileCache().cancelTiles(session);
     if (!canceltiles.empty())
     {
@@ -1439,13 +1654,25 @@ void DocumentBroker::setModified(const bool value)
     if (_isModified != value)
     {
         _isModified = value;
+#ifndef MOBILEAPP
         Admin::instance().modificationAlert(_docKey, getPid(), value);
+#endif
     }
 
     _storage->setUserModified(value);
     _tileCache->setUnsavedChanges(value);
 }
 
+bool DocumentBroker::isInitialSettingSet(const std::string& name) const
+{
+    return _isInitialStateSet.find(name) != _isInitialStateSet.end();
+}
+
+void DocumentBroker::setInitialSetting(const std::string& name)
+{
+    _isInitialStateSet.emplace(name);
+}
+
 bool DocumentBroker::forwardToChild(const std::string& viewId, const std::string& message)
 {
     assertCorrectThread();
@@ -1470,8 +1697,9 @@ bool DocumentBroker::forwardToChild(const std::string& viewId, const std::string
         {
             // The json options must come last.
             msg = tokens[0] + ' ' + tokens[1] + ' ' + tokens[2];
-            msg += " jail=" + _uriJailed.toString() + ' ';
-            msg += Poco::cat(std::string(" "), tokens.begin() + 3, tokens.end());
+            msg += " jail=" + _uriJailed;
+            msg += " xjail=" + _uriJailedAnonym;
+            msg += ' ' + Poco::cat(std::string(" "), tokens.begin() + 3, tokens.end());
         }
 
         _childProcess->sendTextFrame(msg);
@@ -1616,7 +1844,9 @@ void DocumentBroker::broadcastMessage(const std::string& message)
 void DocumentBroker::updateLastActivityTime()
 {
     _lastActivityTime = std::chrono::steady_clock::now();
+#ifndef MOBILEAPP
     Admin::instance().updateLastActivityTime(_docKey);
+#endif
 }
 
 void DocumentBroker::getIOStats(uint64_t &sent, uint64_t &recv)
@@ -1633,18 +1863,6 @@ void DocumentBroker::getIOStats(uint64_t &sent, uint64_t &recv)
     }
 }
 
-ConvertToBroker::~ConvertToBroker()
-{
-    if (!_uriOrig.empty())
-    {
-        // Remove source file and directory
-        Poco::Path path = _uriOrig;
-        Poco::File(path).remove();
-        Poco::File(path.makeParent()).remove();
-        FileUtil::removeFile(_uriOrig);
-    }
-}
-
 void DocumentBroker::dumpState(std::ostream& os)
 {
     std::unique_lock<std::mutex> lock(_mutex);
@@ -1652,7 +1870,9 @@ void DocumentBroker::dumpState(std::ostream& os)
     uint64_t sent, recv;
     getIOStats(sent, recv);
 
-    os << " Broker: " << _filename << " pid: " << getPid();
+    auto now = std::chrono::steady_clock::now();
+
+    os << " Broker: " << LOOLWSD::anonymizeUrl(_filename) << " pid: " << getPid();
     if (_markToDestroy)
         os << " *** Marked to destroy ***";
     else
@@ -1660,20 +1880,22 @@ void DocumentBroker::dumpState(std::ostream& os)
     if (_isLoaded)
         os << "\n  loaded in: " << _loadDuration.count() << "ms";
     else
-        os << "\n  still loading...";
+        os << "\n  still loading... " <<
+            std::chrono::duration_cast<std::chrono::seconds>(
+                now - _threadStart).count() << "s";
     os << "\n  sent: " << sent;
     os << "\n  recv: " << recv;
     os << "\n  modified?: " << _isModified;
     os << "\n  jail id: " << _jailId;
-    os << "\n  filename: " << _filename;
+    os << "\n  filename: " << LOOLWSD::anonymizeUrl(_filename);
     os << "\n  public uri: " << _uriPublic.toString();
-    os << "\n  jailed uri: " << _uriJailed.toString();
+    os << "\n  jailed uri: " << LOOLWSD::anonymizeUrl(_uriJailed);
     os << "\n  doc key: " << _docKey;
     os << "\n  doc id: " << _docId;
     os << "\n  num sessions: " << _sessions.size();
     const std::time_t t = std::chrono::system_clock::to_time_t(
-        std::chrono::system_clock::now()
-        + (_lastSaveTime - std::chrono::steady_clock::now()));
+        std::chrono::time_point_cast<std::chrono::seconds>(
+            std::chrono::system_clock::now() + (_lastSaveTime - now)));
     os << "\n  last saved: " << std::ctime(&t);
     os << "\n  cursor " << _cursorPosX << ", " << _cursorPosY
       << "( " << _cursorWidth << "," << _cursorHeight << ")\n";
