@@ -251,7 +251,7 @@ bool FileServerRequestHandler::isAdminLoggedIn(const HTTPRequest& request,
 
     Poco::Net::HTTPCookie cookie("jwt", jwtToken);
     // bundlify appears to add an extra /dist -> dist/dist/admin
-    cookie.setPath("/loleaflet/dist/");
+    cookie.setPath(LOOLWSD::ServiceRoot + "/loleaflet/dist/");
     cookie.setSecure(LOOLWSD::isSSLEnabled() ||
                      LOOLWSD::isSSLTermination());
     response.addCookie(cookie);
@@ -412,52 +412,52 @@ void FileServerRequestHandler::handleRequest(const HTTPRequest& request, Poco::M
     catch (const Poco::Net::NotAuthenticatedException& exc)
     {
         LOG_ERR("FileServerRequestHandler::NotAuthenticated: " << exc.displayText());
-
-        // Unauthorized.
-        std::ostringstream oss;
-        oss << "HTTP/1.1 401\r\n"
-            << "Date: " << Poco::DateTimeFormatter::format(Poco::Timestamp(), Poco::DateTimeFormat::HTTP_FORMAT) << "\r\n"
-            << "User-Agent: " << WOPI_AGENT_STRING << "\r\n"
-            << "Content-Length: 0\r\n"
-            << "WWW-Authenticate: Basic realm=\"online\"\r\n"
-            << "\r\n";
-        socket->send(oss.str());
+        sendError(401, request, socket, "", "", "WWW-authenticate: Basic realm=\"online\"\r\n");
     }
     catch (const Poco::FileAccessDeniedException& exc)
     {
         LOG_ERR("FileServerRequestHandler: " << exc.displayText());
-
-        // TODO return some 403 page?
-        std::ostringstream oss;
-        oss << "HTTP/1.1 403\r\n"
-            << "Date: " << Poco::DateTimeFormatter::format(Poco::Timestamp(), Poco::DateTimeFormat::HTTP_FORMAT) << "\r\n"
-            << "User-Agent: " << WOPI_AGENT_STRING << "\r\n"
-            << "Content-Length: 0\r\n"
-            << "\r\n";
-        socket->send(oss.str());
+        sendError(403, request, socket, "403 - Access denied!",
+                  "You are unable to access");
     }
     catch (const Poco::FileNotFoundException& exc)
     {
-        LOG_ERR("FileServerRequestHandler: " << exc.displayText());
-
-        // 404 not found
-        std::ostringstream oss;
-        oss << "HTTP/1.1 404\r\n"
-            << "Date: " << Poco::DateTimeFormatter::format(Poco::Timestamp(), Poco::DateTimeFormat::HTTP_FORMAT) << "\r\n"
-            << "User-Agent: " << WOPI_AGENT_STRING << "\r\n"
-            << "Content-Length: 0\r\n"
-            << "\r\n";
-        socket->send(oss.str());
+        LOG_WRN("FileServerRequestHandler: " << exc.displayText());
+        sendError(404, request, socket, "404 - file not found!",
+                  "There seems to be a problem locating");
     }
 }
 
+void FileServerRequestHandler::sendError(int errorCode, const Poco::Net::HTTPRequest& request,
+                                         const std::shared_ptr<StreamSocket>& socket,
+                                         const std::string& shortMessage, const std::string& longMessage,
+                                         const std::string& extraHeader)
+{
+    Poco::URI requestUri(request.getURI());
+    const std::string& path = requestUri.getPath();
+    std::ostringstream oss;
+    oss << "HTTP/1.1 " << errorCode << "\r\n"
+        << "Content-Type: text/html charset=UTF-8\r\n"
+        << "Date: " << Poco::DateTimeFormatter::format(Poco::Timestamp(), Poco::DateTimeFormat::HTTP_FORMAT) << "\r\n"
+        << "User-Agent: " << WOPI_AGENT_STRING << "\r\n"
+        << extraHeader
+        << "\r\n";
+    if (!shortMessage.empty())
+    {
+        oss << "<h1>Error: " << shortMessage << "</h1>"
+            << "<p>" << longMessage << " " << path << "</p>"
+            << "<p>Please contact your system administrator.</p>";
+    }
+    socket->send(oss.str());
+}
+
 void FileServerRequestHandler::readDirToHash(const std::string &basePath, const std::string &path)
 {
     struct dirent *currentFile;
     struct stat fileStat;
     DIR *workingdir;
 
-    LOG_TRC("Pre-reading directory: " << basePath << path << "\n");
+    LOG_TRC("Pre-reading directory: " << basePath << path);
     workingdir = opendir((basePath + path).c_str());
 
     while ((currentFile = readdir(workingdir)) != nullptr)
@@ -473,7 +473,7 @@ void FileServerRequestHandler::readDirToHash(const std::string &basePath, const
 
         else if (S_ISREG(fileStat.st_mode))
         {
-            LOG_TRC("Reading file: '" << basePath << relPath << " as '" << relPath << "'\n");
+            LOG_TRC("Reading file: '" << basePath << relPath << " as '" << relPath << "'");
 
             std::ifstream file(basePath + relPath, std::ios::binary);
 
@@ -488,7 +488,8 @@ void FileServerRequestHandler::readDirToHash(const std::string &basePath, const
             compressedFile.reserve(fileStat.st_size);
             std::string uncompressedFile;
             uncompressedFile.reserve(fileStat.st_size);
-            do {
+            do
+            {
                 file.read(&buf[0], fileStat.st_size);
                 const long unsigned int size = file.gcount();
                 if (size == 0)
@@ -599,7 +600,7 @@ void FileServerRequestHandler::preprocessFile(const HTTPRequest& request, Poco::
             {
                 tokenTtl = std::stoul(accessTokenTtl);
             }
-            catch(const std::exception& exc)
+            catch (const std::exception& exc)
             {
                 LOG_ERR("access_token_ttl must be represented as the number of milliseconds since January 1, 1970 UTC, when the token will expire");
             }
@@ -615,12 +616,13 @@ void FileServerRequestHandler::preprocessFile(const HTTPRequest& request, Poco::
     Poco::replaceInPlace(preprocess, std::string("%ACCESS_HEADER%"), escapedAccessHeader);
     Poco::replaceInPlace(preprocess, std::string("%HOST%"), host);
     Poco::replaceInPlace(preprocess, std::string("%VERSION%"), std::string(LOOLWSD_VERSION_HASH));
+    Poco::replaceInPlace(preprocess, std::string("%SERVICE_ROOT%"), LOOLWSD::ServiceRoot);
 
-    static const std::string linkCSS("<link rel=\"stylesheet\" href=\"/loleaflet/" LOOLWSD_VERSION_HASH "/%s.css\">");
-    static const std::string scriptJS("<script src=\"/loleaflet/" LOOLWSD_VERSION_HASH "/%s.js\"></script>");
+    static const std::string linkCSS("<link rel=\"stylesheet\" href=\"%s/loleaflet/" LOOLWSD_VERSION_HASH "/%s.css\">");
+    static const std::string scriptJS("<script src=\"%s/loleaflet/" LOOLWSD_VERSION_HASH "/%s.js\"></script>");
 
-    std::string brandCSS(Poco::format(linkCSS, std::string(BRANDING)));
-    std::string brandJS(Poco::format(scriptJS, std::string(BRANDING)));
+    std::string brandCSS(Poco::format(linkCSS, LOOLWSD::ServiceRoot, std::string(BRANDING)));
+    std::string brandJS(Poco::format(scriptJS, LOOLWSD::ServiceRoot, std::string(BRANDING)));
 
     const auto& config = Application::instance().config();
 #if ENABLE_SUPPORT_KEY
@@ -628,17 +630,26 @@ void FileServerRequestHandler::preprocessFile(const HTTPRequest& request, Poco::
     SupportKey key(keyString);
     if (!key.verify() || key.validDaysRemaining() <= 0)
     {
-        brandCSS = Poco::format(linkCSS, std::string(BRANDING_UNSUPPORTED));
-        brandJS = Poco::format(scriptJS, std::string(BRANDING_UNSUPPORTED));
+        brandCSS = Poco::format(linkCSS, LOOLWSD::ServiceRoot, std::string(BRANDING_UNSUPPORTED));
+        brandJS = Poco::format(scriptJS, LOOLWSD::ServiceRoot, std::string(BRANDING_UNSUPPORTED));
     }
-#elif ENABLE_DEBUG
-    brandCSS = "";
-    brandJS = "";
 #endif
 
     Poco::replaceInPlace(preprocess, std::string("<!--%BRANDING_CSS%-->"), brandCSS);
     Poco::replaceInPlace(preprocess, std::string("<!--%BRANDING_JS%-->"), brandJS);
 
+    // Customization related to document signing.
+    std::string documentSigningDiv;
+    std::string documentContainerTop("<div id=\"document-container\">");
+    const auto documentSigning = config.getBool("per_document.document_signing", false);
+    if (documentSigning)
+    {
+        documentSigningDiv = "<div id=\"document-signing-bar\"></div>";
+        documentContainerTop = "<div id=\"document-container\" style=\"top: 100px;\">";
+    }
+    Poco::replaceInPlace(preprocess, std::string("<!--%DOCUMENT_SIGNING_DIV%-->"), documentSigningDiv);
+    Poco::replaceInPlace(preprocess, std::string("<!--%DOCUMENT_CONTAINER%-->"), documentContainerTop);
+
     const auto loleafletLogging = config.getString("loleaflet_logging", "false");
     Poco::replaceInPlace(preprocess, std::string("%LOLEAFLET_LOGGING%"), loleafletLogging);
     const std::string outOfFocusTimeoutSecs= config.getString("per_view.out_of_focus_timeout_secs", "60");
@@ -670,53 +681,49 @@ void FileServerRequestHandler::preprocessFile(const HTTPRequest& request, Poco::
            << "font-src 'self' data:; "
            << "object-src blob:; ";
 
-    std::string frameAncestor;
-    const auto it = request.find("Referer"); // Referer[sic]
-    if (it != request.end())
+#if !ENABLE_DEBUG // always allow iframe embedding in debug mode
+    // Frame ancestors: Allow loolwsd host, wopi host and anything configured.
+    std::string configFrameAncestor = config.getString("net.frame_ancestors", "");
+    std::string frameAncestors = configFrameAncestor;
+    Poco::URI uriHost(host);
+    if (uriHost.getHost() != configFrameAncestor)
+        frameAncestors += " " + uriHost.getHost() + ":*";
+
+    for (const auto& param : params)
     {
-        frameAncestor = it->second;
-        LOG_TRC("Picking frame ancestor from HTTP Referer header: " << frameAncestor);
-    }
-    else // Use WOPISrc value if Referer is absent
-    {
-        for (const auto& param : params)
+        if (param.first == "WOPISrc")
         {
-            if (param.first == "WOPISrc")
+            std::string wopiFrameAncestor;
+            Poco::URI::decode(param.second, wopiFrameAncestor);
+            Poco::URI uriWopiFrameAncestor(wopiFrameAncestor);
+            // Remove parameters from URL
+            wopiFrameAncestor = uriWopiFrameAncestor.getHost();
+            if (wopiFrameAncestor != uriHost.getHost() && wopiFrameAncestor != configFrameAncestor)
             {
-                Poco::URI::decode(param.second, frameAncestor);
-                LOG_TRC("Picking frame ancestor from WOPISrc: " << frameAncestor);
-                break;
+                frameAncestors += " " + wopiFrameAncestor + ":*";
+                LOG_TRC("Picking frame ancestor from WOPISrc: " << wopiFrameAncestor);
             }
+            break;
         }
     }
 
-    // Keep only the origin, reject everything else
-    Poco::URI uriFrameAncestor(frameAncestor);
-    const std::string& frameAncestorScheme = uriFrameAncestor.getScheme();
-    const std::string& frameAncestorHost = uriFrameAncestor.getHost();
-
-    if (!frameAncestor.empty() && Util::isValidURIScheme(frameAncestorScheme) && Util::isValidURIHost(frameAncestorHost))
+    if (!frameAncestors.empty())
     {
-        frameAncestor = frameAncestorScheme + "://" + frameAncestorHost + ":" + std::to_string(uriFrameAncestor.getPort());
-
-        LOG_TRC("Final frame ancestor: " << frameAncestor);
-
-        // Replaced by frame-ancestors in CSP but some oldies don't know about that
-        oss << "X-Frame-Options: allow-from " << frameAncestor << "\r\n";
-        cspOss << "img-src 'self' data: " << frameAncestor << "; "
-               << "frame-ancestors " << frameAncestor;
+        LOG_TRC("Allowed frame ancestors: " << frameAncestors);
+        // X-Frame-Options supports only one ancestor, ignore that
+        //(it's deprecated anyway and CSP works in all major browsers)
+        cspOss << "img-src 'self' data: " << frameAncestors << "; "
+                << "frame-ancestors " << frameAncestors;
     }
     else
     {
-        LOG_TRC("Denied frame ancestor: " << frameAncestor);
-
-        cspOss << "img-src 'self' data: ;";
-        oss << "X-Frame-Options: deny\r\n";
+        LOG_TRC("Denied all frame ancestors");
+        cspOss << "img-src 'self' data: none;";
     }
-
     cspOss << "\r\n";
     // Append CSP to response headers too
     oss << cspOss.str();
+#endif
 
     // Setup HTTP Public key pinning
     if ((LOOLWSD::isSSLEnabled() || LOOLWSD::isSSLTermination()) && config.getBool("ssl.hpkp[@enable]", false))
@@ -792,13 +799,13 @@ void FileServerRequestHandler::preprocessAdminFile(const HTTPRequest& request,co
     if (!FileServerRequestHandler::isAdminLoggedIn(request, response))
         throw Poco::Net::NotAuthenticatedException("Invalid admin login");
 
-    static const std::string scriptJS("<script src=\"/loleaflet/dist/%s.js\"></script>");
+    static const std::string scriptJS("<script src=\"%s/loleaflet/" LOOLWSD_VERSION_HASH "/%s.js\"></script>");
     static const std::string footerPage("<div class=\"footer navbar-fixed-bottom text-info text-center\"><strong>Key:</strong> %s &nbsp;&nbsp;<strong>Expiry Date:</strong> %s</div>");
 
     const std::string relPath = getRequestPathname(request);
     LOG_DBG("Preprocessing file: " << relPath);
     std::string adminFile = *getUncompressedFile(relPath);
-    std::string brandJS(Poco::format(scriptJS, std::string(BRANDING)));
+    std::string brandJS(Poco::format(scriptJS, LOOLWSD::ServiceRoot, std::string(BRANDING)));
     std::string brandFooter;
 
 #if ENABLE_SUPPORT_KEY
@@ -811,12 +818,12 @@ void FileServerRequestHandler::preprocessAdminFile(const HTTPRequest& request,co
         brandJS = Poco::format(scriptJS, std::string(BRANDING_UNSUPPORTED));
         brandFooter = Poco::format(footerPage, key.data(), Poco::DateTimeFormatter::format(key.expiry(), Poco::DateTimeFormat::RFC822_FORMAT));
     }
-#elif ENABLE_DEBUG
-    brandJS = "";
 #endif
 
     Poco::replaceInPlace(adminFile, std::string("<!--%BRANDING_JS%-->"), brandJS);
     Poco::replaceInPlace(adminFile, std::string("<!--%FOOTER%-->"), brandFooter);
+    Poco::replaceInPlace(adminFile, std::string("%VERSION%"), std::string(LOOLWSD_VERSION_HASH));
+    Poco::replaceInPlace(adminFile, std::string("%SERVICE_ROOT%"), LOOLWSD::ServiceRoot);
 
     // Ask UAs to block if they detect any XSS attempt
     response.add("X-XSS-Protection", "1; mode=block");
